from pathlib import Path, PosixPath

import numpy as np
# from mne.io.brainvision.brainvision import RawBrainVision
# from mne.io.edf.edf import RawEDF, RawGDF
# from mne.io.eeglab.eeglab import RawEEGLAB

from .log import logger
from .recording import Recording
from .Event import Event
from .RecordingMetadata import RecordingMetadata
from .constants import EEG_MONTAGES
from .errors import BreesyError, BreesyInternalError, protect_from_lib_error
from .load_events import find_and_load_event_for_file
from .type_hints import enforce_type_hints

# Note: if changing this, then update load_dataset if/else statement as well
# SUPPORTED_FILE_EXTENSIONS = ['.gdf', '.bdf', '.eeg', '.csv', '.mat', '.tsv', '.vhdr', '.npy']
SUPPORTED_FILE_EXTENSIONS = ['.csv', '.mat', '.tsv', '.npy', '.eeg', '.vhdr', '.avg', '.seg']


@enforce_type_hints
def load_recording(filename: str | Path | PosixPath, sample_rate: int | float | None = None) -> Recording:
    """Load a recording from a file with a supported extension.

    :param filename: Path to the file to load
    :param sample_rate: Optional sampling rate in Hz

    :return: Recording object containing the loaded data
    """
    filename = str(filename)
    file_extension = Path(filename).suffix
    if not file_extension:
        raise BreesyError("File name should contain an extension",
                          "Provide a file with a supported extension, example: 'my_data.gdf'")

    is_supported_extension = file_extension in SUPPORTED_FILE_EXTENSIONS
    if not is_supported_extension:
        raise BreesyError(f"Unsupported file extension: {file_extension}",
                          "Provide a file with a supported extension, example: 'my_data.gdf'. "
                          f"Currently supported extensions: {SUPPORTED_FILE_EXTENSIONS}")

    path_filename = Path(filename)
    is_valid_filename = Path.is_file(path_filename)
    if not is_valid_filename:
        raise BreesyError(
            f"File not found: {filename}", "a) Make sure that the file is in your working directory; "
                                           f"the current working directory is: {Path.cwd()}. "
                                           f"b) Double check that your file explorer is showing the full file name "
                                           f"- Windows does not show file extensions by default.")

    # if file_extension == '.gdf':
    #     return load_gdf(filename)
    # elif file_extension == '.bdf':
    #     return load_bdf(filename)
    elif file_extension in ['.eeg', '.avg', '.seg', '.vhdr']:
        from .load_brainvision import load_brainvision_eeg
        return load_brainvision_eeg(filename)
    elif file_extension == '.csv':
        return load_csv(filename)
    elif file_extension == '.mat':
        return load_mat(filename, sample_rate=sample_rate)
    elif file_extension == '.tsv':
        return load_tsv(filename)
    elif file_extension == '.npy':
        return load_numpy(filename)
    else:
        raise BreesyInternalError("File extension valid but not handled")


# @enforce_type_hints
# def load_gdf(filename: str):
#     """Load a GDF file.

#     :param filename: Path to the GDF file

#     :return: Recording object containing the loaded data
#     """
#     raw = _read_raw_gdf(filename)
#     return _convert_mne_to_recording(raw, filename)

# @enforce_type_hints
# def load_bdf(filename: str):
#     """Load a BDF file.

#     :param filename: Path to the BDF file

#     :return: Recording object containing the loaded data
#     """
#     raw = _read_raw_bdf(filename)
#     return _convert_mne_to_recording(raw, filename)

# @enforce_type_hints
# def load_eeglab_eeg(filename: str):
#     """Load an EEGLAB EEG file.

#     :param filename: Path to the EEG file

#     :return: Recording object containing the loaded data
#     """
#     raw = _read_raw_eeglab(filename)
#     return _convert_mne_to_recording(raw, filename)


def _mat_extract_meaningful_keys(mat_data: dict) -> list[str]:
    keys = [k for k in mat_data.keys() if not k.startswith('__')]
    if not keys:
        raise BreesyError(
            "No valid data found. The MATLAB file appears to be empty or contains only metadata."
        )
    elif len(keys) == 1:
        logger.warning(
            "data file likely contains only the recording and "
            "no useful metadata. Please find the sample rate and channel names "
            "in your data description."
        )  # TODO: fast instructions or link to documentation
    return keys


@enforce_type_hints
def load_mat(filename: str, sample_rate: int | float | None = None) -> Recording:
    """Load a MATLAB .mat file.

    :param filename: Path to the .mat file
    :param sample_rate: Sampling rate in Hz

    :return: Recording object containing the loaded data
    """
    mat_data = _loadmat(filename)

    meaningful_keys = _mat_extract_meaningful_keys(mat_data)

    data_key = None

    if len(meaningful_keys) == 1:
        data_key = meaningful_keys[0]
    else:
        # Find the largest array that looks like EEG data
        largest_size = 0
        for key in meaningful_keys:
            value = mat_data[key]
            if isinstance(value, np.ndarray) and value.size > largest_size and value.ndim >= 2:
                largest_size = value.size
                data_key = key

    data = mat_data[data_key]  # TODO: there can be 'O' types, as AI previously suggested?
    if not isinstance(data, np.ndarray):
        raise BreesyError(
            f"Problem reading data which is of type {type(data)}. It may be that Breesy failed to read the file correctly."  # TODO: test different files
        )
    if data.ndim != 2:
        raise BreesyError(
            "Found data has an incorrect number of dimensions. "
            "Expected data.ndim to be 2, but got "
            f"{data.ndim} dimensions instead.",
            "The data may be not EEG data, "
            "there are several records joined into one file, "
            "or Breesy did not found the correct data array in the file."
        )

    if data.shape[0] > data.shape[1]:
        logger.warning("The first data dimension seems to be the samples, because data contains more rows than columns. Dimensions will be switched places.")
        data = data.T
    n_channels, n_samples = data.shape
    default_channel_names = [f'Channel_{i + 1}' for i in range(data.shape[0])]

    # find or create channel names
    channel_name_keys = _find_keys_in_data(meaningful_keys, ['ch', 'ch name', 'ch names', 'channel names', 'channels', 'names', '10-20', '10-10', 'pos', 'positions'])
    if len(channel_name_keys) == 0:
        channel_names = default_channel_names
        channel_name_key = None
    elif len(channel_name_keys) > 1:
        logger.warning(
            "found several keys in data which may contain channel names: "
            f"{channel_name_keys}. "
            "Using the first one.")
        channel_name_key = channel_name_keys[0]
    else:
        channel_name_key = channel_name_keys[0]

    if channel_name_key:
        found_names = mat_data[channel_name_key]
        if not isinstance(found_names, (list, tuple, np.ndarray)):
            logger.warning(
                "found channel name information does not appear to be a list: "
                f"{found_names}. "
                "Default channel names will be used instead.")
            channel_names = default_channel_names
        elif (n_channels != len(found_names)) and (n_samples != len(found_names)):
            logger.warning(
                f"the number of channel names ({len(found_names)}) provided does not match the "
                f"number of channels ({n_channels}) in the EEG data. "
                "Default channel names will be used instead."
            )
            channel_names = default_channel_names
        elif (n_channels != len(channel_names)) and (n_samples == len(channel_names)):
            logger.warning('data rows and columns will be switched places according to the number of found channel names.')
            data = data.T
            n_channels, n_samples = data.shape
            channel_names = found_names
        else:
            channel_names = found_names

    if sample_rate:
        logger.info(f"Overwriting sample rate to {sample_rate}")
        sample_rate = float(sample_rate)
    else:
        # try to find it in keys
        sample_rate = _detect_sample_rate_in_mat_keys(mat_data, meaningful_keys)
        if not sample_rate:
            # try to calculate it from time channel, if present
            sample_rate = _detect_sample_rate_from_time_column(data)
            if not sample_rate:
                raise BreesyError(
                    "No info found about sample rate.",
                    "Please find the recording's sample rate in data description and set manually by "
                    "passing an additional parameter 'sample_rate' into the data loading function. "
                    "Example: load_recording(path, sample_rate=128)."
                )

    data, channel_names = _leave_only_eeg_channels(data, channel_names)
    # channel names may not start from 1 after this, but maybe we should leave it like this

    metadata = RecordingMetadata(
        name=Path(filename).stem,
        file_path=Path(filename),
        description=f"MATLAB data loaded from {filename}"
    )

    return Recording(
        data=data,
        channel_names=channel_names,
        sample_rate=sample_rate,
        events=None,  # No events for now - could be added as a separate function
        metadata=metadata
    )


@enforce_type_hints
def load_delimited_text(filename: str, delimiter: str = ',', sample_rate: int | float | None = None) -> Recording:
    """Load a delimited text file (CSV, TSV, etc.).

    :param filename: Path to the text file
    :param delimiter: Character used to separate values in the file
    :param sample_rate: Sampling rate in Hz. If not provided, will attempt to guess from time column. If sample rate cannot be guessed, an error will be raised.

    :return: Recording object containing the loaded data
    """
    import csv

    # First pass: read headers and determine structure
    with open(filename, 'r', newline='') as file:
        reader = csv.reader(file, delimiter=delimiter)
        headers = next(reader)  # Get column names

        # Check if first column might be time
        time_column_idx = None
        if headers[0].lower() in ['time', 't', 'timestamp']:
            time_column_idx = 0

        # Exclude time column from channel names if present
        if time_column_idx is not None:
            channel_names = headers[1:]
        else:
            channel_names = headers

        # Count rows for pre-allocation
        row_count = sum(1 for _ in reader)

    # Second pass: read all data
    with open(filename, 'r', newline='') as file:
        reader = csv.reader(file, delimiter=delimiter)
        next(reader)  # Skip header

        # Pre-allocate data array
        data = np.zeros((len(channel_names), row_count))
        time_values = None

        if time_column_idx is not None:
            time_values = np.zeros(row_count)

        # Fill data
        for i, row in enumerate(reader):
            if i >= row_count:
                break

            # Convert values to float, handling potential errors
            try:
                row_values = [float(val) for val in row]
            except ValueError:
                # Skip rows that can't be converted to float
                continue

            if time_column_idx is not None:
                time_values[i] = row_values[time_column_idx]
                data_values = row_values[:time_column_idx] + row_values[time_column_idx + 1:]
            else:
                data_values = row_values

            for j, val in enumerate(data_values):
                if j < len(channel_names):
                    data[j, i] = val

    if sample_rate is not None:
        sample_rate = float(sample_rate)
    else:
        if time_values is not None and len(time_values) > 1:
            # Estimate sample rate
            time_diffs = np.diff(time_values)
            avg_diff = float(np.mean(time_diffs))
            if avg_diff > 0:
                sample_rate = float(round(1.0 / avg_diff))
        
        if sample_rate is None:
            raise BreesyError(
                "Sample rate was not provided and it could not be guessed from the data. "
                "Please provide the sample rate manually by passing the 'sample_rate' parameter.",
                "Provide the sample rate as a named argument to the function, e.g.: load_delimited_text('file.csv', sample_rate=500)"
            )

    # Create metadata
    metadata = RecordingMetadata(
        name=Path(filename).stem,
        file_path=Path(filename),
        description=f"Data loaded from {filename}"
    )

    # Look for event columns (columns with mostly 0s and few 1s)
    events = []
    for i, col_name in enumerate(channel_names):
        col_data = data[i, :]
        unique_vals = set(np.unique(col_data))
        # Check if column is binary and sparse (potential event marker)
        if unique_vals <= {0, 1} and np.mean(col_data) < 0.1:
            event_indices = np.where(col_data == 1)[0]
            for idx in event_indices:
                events.append(Event(name=col_name, index=idx))

    return Recording(
        data=data,
        channel_names=channel_names,
        sample_rate=sample_rate,
        events=events,
        metadata=metadata
    )

# Wrapper functions for backward compatibility
@enforce_type_hints
def load_csv(filename: str) -> Recording:
    """Load a CSV file.

    :param filename: Path to the CSV file

    :return: Recording object containing the loaded data
    """
    return load_delimited_text(filename, delimiter=',')


@enforce_type_hints
def load_tsv(filename: str) -> Recording:
    """Load a TSV file.

    :param filename: Path to the TSV file

    :return: Recording object containing the loaded data
    """
    return load_delimited_text(filename, delimiter='\t')


# Note: mostly LLM generated, looks reasonable
@enforce_type_hints
def load_numpy(filename: str, sample_rate: int | float | None = None) -> Recording:
    """Load a numpy .npy file.

    :param filename: Path to the .npy file
    :param sample_rate: Sampling rate in Hz. If None, defaults to 1000 Hz.

    :return: Recording object containing the loaded data
    """
    # Load the NumPy array with error protection
    data = _np_load(filename)

    # TODO: #12 we need to handle this case properly, currently just taking first trial
    if data.ndim == 3:
        # likely many recordings instead of one
        data = data[0]

    # Ensure it's a 2D array
    if data.ndim != 2:
        raise BreesyError(
            f"The NumPy array has {data.ndim} dimensions, but 2 dimensions are required.",
            "Make sure the NumPy file contains a 2D array with shape (channels, samples)."
        )

    # If the first dimension is larger, assume it's (samples, channels) and transpose
    if data.shape[0] > data.shape[1]:
        logger.warning("The first data dimension seems to be the samples, because data contains more rows than columns. Dimensions will be switched places.")
        data = data.T

    # Create default channel names
    channel_names = [f'Channel_{i+1}' for i in range(data.shape[0])]

    if sample_rate is None:
        logger.warning("No sample rate provided. Using default of 1000 Hz.")
        sample_rate = 1000
    else:
        sample_rate = float(sample_rate)

    # Create metadata
    metadata = RecordingMetadata(
        name=Path(filename).stem,
        file_path=Path(filename),
        description=f"NumPy data loaded from {filename}"
    )

    # Find and load events if available
    events = find_and_load_event_for_file(filename, sample_rate)

    return Recording(
        data=data,
        channel_names=channel_names,
        sample_rate=sample_rate,
        events=events,
        metadata=metadata
    )


@enforce_type_hints
def from_array(data: np.ndarray, sample_rate: int | float | None = None, channel_names: list[str] | None = None) -> Recording:
    """Convert a numpy array into a Recording object.

    :param data: NumPy array containing EEG data. Should be 2D with shape (channels, samples) or (samples, channels).
                 If a time column is detected, it will be used to extract the sample rate and removed from the data.
    :param sample_rate: Sampling rate in Hz. If not provided, will attempt to detect from time column.
                        If no time column is found and sample_rate is None, an error will be raised.
    :param channel_names: List of channel names. If not provided, default names will be generated.

    :return: Recording object containing the loaded data
    """
    # Ensure it's a 2D array
    if data.ndim != 2:
        raise BreesyError(
            f"The NumPy array has {data.ndim} dimensions, but 2 dimensions are required.",
            "Make sure the NumPy array is 2D with shape (channels, samples) or (samples, channels)."
        )

    # Auto-transpose if the first dimension is larger (likely samples, channels)
    if data.shape[0] > data.shape[1]:
        logger.warning("The first data dimension seems to be the samples, because data contains more rows than columns. Dimensions will be switched places.")
        data = data.T

    # Check for time column and extract sample rate if needed
    time_column_idx = None
    extracted_sample_rate = None

    for i in range(data.shape[0]):
        if _is_likely_time_column(data[i]):
            time_column_idx = i
            time_step = float(data[i, 1] - data[i, 0])
            extracted_sample_rate = float(1 / time_step)
            logger.info(f"Detected time column at index {i}. Extracted sample rate: {extracted_sample_rate} Hz")
            break

    # Remove time column from data if found
    if time_column_idx is not None:
        data = np.delete(data, time_column_idx, axis=0)

    # Determine final sample rate
    if sample_rate is not None:
        if extracted_sample_rate is not None and abs(sample_rate - extracted_sample_rate) > 0.01:
            logger.warning(f"Warning: Provided sample_rate ({sample_rate} Hz) differs from extracted sample rate ({extracted_sample_rate} Hz). Using the provided sample_rate.")
        sample_rate = float(sample_rate)
    elif extracted_sample_rate is not None:
        sample_rate = extracted_sample_rate
    else:
        raise BreesyError(
            "No sample rate provided and could not detect sample rate from data.",
            "Please provide the sample rate manually by passing the 'sample_rate' parameter. "
            "Example: from_array(data, sample_rate=250)"
        )

    # Create channel names
    if channel_names is None:
        channel_names = [f'Channel_{i+1}' for i in range(data.shape[0])]
    elif len(channel_names) != data.shape[0]:
        raise BreesyError(
            f"The number of channel names ({len(channel_names)}) does not match the number of channels ({data.shape[0]}) in the data.",
            "Make sure that the number of channel names matches the number of channels in the data. "
            f"Expected {data.shape[0]} channel names."
        )

    # Create metadata
    metadata = RecordingMetadata(
        name="NumPy Recording",
        file_path=None,
        description="Recording created from NumPy array"
    )

    return Recording(
        data=data,
        channel_names=channel_names,
        sample_rate=sample_rate,
        events=None,
        metadata=metadata
    )

# # Note: LLM generated, looks reasonable
# @enforce_type_hints
# def load_brainvision_eeg(filename: str) -> Recording:
#     """Load a BrainVision EEG file.

#     :param filename: Path to the BrainVision file (.vhdr)

#     :return: Recording object containing the loaded data
#     """
#     file_path = Path(filename)

#     # Ensure the file exists
#     if not file_path.exists():
#         raise BreesyError(
#             f"File not found: {filename}", "a) Make sure that the file is in your working directory; "
#                                            f"the current working directory is: {Path.cwd()}. "
#                                            f"b) Double check that your file explorer is showing the full file name "
#                                            f"- Windows does not show file extensions by default.")

#     # Determine which file we're dealing with
#     if file_path.suffix.lower() == '.eeg':
#         # If .eeg file provided, look for the corresponding .vhdr
#         vhdr_path = file_path.with_suffix('.vhdr')
#         if not vhdr_path.exists():
#             raise BreesyError(
#                 f"Header file not found for {filename}",
#                 "BrainVision files require a .vhdr file. Make sure it's in the same directory as the .eeg file."
#             )
#         filename = str(vhdr_path)
#     elif file_path.suffix.lower() != '.vhdr':
#         raise BreesyError(
#             f"Unsupported BrainVision file extension: {file_path.suffix}",
#             "Please provide a .vhdr or .eeg file."
#         )

#     # Load the EEG data using MNE
#     raw = _read_raw_brainvision(filename)

#     return _convert_mne_to_recording(raw, filename)

# @enforce_type_hints
# def load_all_files_in_folder_by_extension(file_extension: str, folder_path: str = ""):
#     """Load all files with a specific extension from a folder.

#     :param file_extension: File extension to filter by (e.g., '.mat')
#     :param folder_path: Path to the folder containing the files

#     :return: List of Recording objects containing the loaded data
#     """
#     if not isinstance(folder_path, str):
#         raise BreesyError("Invalid folder_path argument type",
#                           "Provide a string as the folder path, example: 'data_folder', or '' (empty string or period) for current working directory.")

#     if not isinstance(file_extension, str):
#         raise BreesyError("Invalid file_extension argument type",
#                           "Provide a string as the file extension, example: '.gdf'")

#     folder = Path(folder_path)
#     if not folder.is_dir():
#         raise BreesyError(f"The folder path was not found: {folder.resolve()}",
#                           f"Check the folder path manually to make sure that the folder actually exists.")

#     full_path = folder.resolve()

#     files = list(full_path.glob(f"*{file_extension}"))
#     if not files:
#         raise BreesyError(
#             f"No {file_extension} files found in folder: '{folder_path}' (resolved absolute path: {folder.resolve()})",
#             f"Check the folder to make sure that it actually contains {file_extension} files.")

#     data_list = []
#     for file in files:
#         data = load_dataset(str(file))
#         data_list.append(data)

#     return data_list


# @enforce_type_hints
# def load_all_mat_files_in_folder(folder_path: str = ""):
#     """Load all MATLAB .mat files from a folder.

#     :param folder_path: Path to the folder containing the files

#     :return: List of Recording objects containing the loaded data
#     """
#     return load_all_files_in_folder_by_extension(".mat", folder_path)


# @enforce_type_hints
# def load_all_gdf_files_in_folder(folder_path: str = ""):
#     """Load all GDF files from a folder.

#     :param folder_path: Path to the folder containing the files

#     :return: List of Recording objects containing the loaded data
#     """
#     return load_all_files_in_folder_by_extension(".gdf", folder_path)


# @enforce_type_hints
# def load_all_bdf_files_in_folder(folder_path: str = ""):
#     """Load all BDF files from a folder.

#     :param folder_path: Path to the folder containing the files

#     :return: List of Recording objects containing the loaded data
#     """
#     return load_all_files_in_folder_by_extension(".bdf", folder_path)


# @enforce_type_hints
# def load_all_eeg_files_in_folder(folder_path: str = ""):
#     """Load all EEG files from a folder.

#     :param folder_path: Path to the folder containing the files

#     :return: List of Recording objects containing the loaded data
#     """
#     return load_all_files_in_folder_by_extension(".eeg", folder_path)


# @enforce_type_hints
# def load_all_csv_files_in_folder(folder_path: str = ""):
#     """Load all CSV files from a folder.

#     :param folder_path: Path to the folder containing the files

#     :return: List of Recording objects containing the loaded data
#     """
#     return load_all_files_in_folder_by_extension(".csv", folder_path)


# @enforce_type_hints
# def load_all_tsv_files_in_folder(folder_path: str = ""):
#     """Load all TSV files from a folder.

#     :param folder_path: Path to the folder containing the files

#     :return: List of Recording objects containing the loaded data
#     """
#     return load_all_files_in_folder_by_extension(".tsv", folder_path)

# @enforce_type_hints
# def load_all_brainvision_files_in_folder(folder_path: str = ""):
#     """Load all BrainVision files from a folder.

#     :param folder_path: Path to the folder containing the files

#     :return: List of Recording objects containing the loaded data
#     """
#     return load_all_files_in_folder_by_extension(".eeg", folder_path)


# @enforce_type_hints
# def load_all_numpy_files_in_folder(folder_path: str = ""):
#     """Load all NumPy .npy files from a folder.

#     :param folder_path: Path to the folder containing the files

#     :return: List of Recording objects containing the loaded data
#     """
#     return load_all_files_in_folder_by_extension(".npy", folder_path)

# --------- Helper functions ---------

@protect_from_lib_error("numpy")
def _np_load(filename):
    return np.load(filename)

# @protect_from_lib_error("mne")
# def _read_raw_gdf(filename):
#     from mne.io import read_raw_gdf
#     return read_raw_gdf(filename, verbose=False)

# @protect_from_lib_error("mne")
# def _read_raw_bdf(filename):
#     from mne.io import read_raw_bdf
#     return read_raw_bdf(filename, verbose=False)

# @protect_from_lib_error("mne")
# def _read_raw_eeglab(filename):
#     from mne.io import read_raw_eeglab
#     return read_raw_eeglab(filename, verbose=False)

# @protect_from_lib_error("mne")
# def _read_raw_brainvision(filename: str):
#     from mne.io import read_raw_brainvision
#     return read_raw_brainvision(filename, verbose=False)

@protect_from_lib_error("scipy")
def _loadmat(filename):
    from scipy.io import loadmat
    return loadmat(filename, squeeze_me=True)

# def _convert_mne_to_recording(raw: RawGDF | RawEDF | RawEEGLAB | RawBrainVision, filename):
#     data = raw.get_data()
#     sample_rate = int(raw.info['sfreq'])
#     channel_names = raw.ch_names

#     # Extract events if available
#     events = []
#     try:
#         # Use MNE to find events
#         from mne import find_events
#         mne_events = find_events(raw)
#         for event in mne_events:
#             sample_idx = event[0]
#             event_id = event[2]
#             event_name = raw.event_id.get(event_id, f"Event_{event_id}")
#             events.append(Event(name=event_name, index=sample_idx))
#     except Exception as e:
#         pass
#     if not events:
#         print("Events not found through mne.find_events")

#     # Try from annotations
#     if not events:
#         try:
#             from mne import events_from_annotations
#             mne_events, event_id = events_from_annotations(raw)
#             # Map event IDs to event names
#             id_to_name = {v: k for k, v in event_id.items()}

#             for event in mne_events:
#                 sample_idx = event[0]
#                 event_code = event[2]
#                 event_name = id_to_name.get(event_code, f"Event_{event_code}")
#                 events.append(Event(name=event_name, index=sample_idx))
#         except Exception as e:
#             pass
#     if not events:
#         print("Events not found through mne.events_from_annotations")

#     # Extract events for peer files
#     # TODO: maybe this needs to be used in ALL load functions
#     if not events:
#         try:
#             events = find_and_load_event_for_file(filename, sample_rate)
#         except Exception as e:
#             pass
#     if not events:
#         print("Events not found in related files")

#     metadata = RecordingMetadata(
#         name=Path(filename).stem,
#         file_path=Path(filename),
#         description=f"Data loaded from {filename}"
#     )

#     return Recording(
#         data=data,
#         channel_names=channel_names,
#         sample_rate=sample_rate,
#         events=events,
#         metadata=metadata
#     )


def _detect_sample_rate_in_mat_keys(mat_data: dict, keys: list[str]) -> float | None:
    sample_rate_keys = _find_keys_in_data(keys, ['sample rate', 'sp', 'sr', 'hz', 'sampling rate', 'sfreq', 'fs', 'srate'])
    if len(sample_rate_keys) == 0:
        return None

    sample_rate_key = sample_rate_keys[0]
    if len(sample_rate_keys) > 1:
        logger.warning(
            "found several keys in data which may contain sample rate:"
            f"\n{sample_rate_keys}\n"
            "Using the first one.")

    sr_value = mat_data[sample_rate_key]
    try:
        if isinstance(sr_value, (int, float, str)):
            return float(sr_value)
        elif isinstance(sr_value, np.ndarray) and sr_value.size == 1:
            return float(sr_value.item())
        elif isinstance(sr_value, (list, tuple)) and len(sr_value) == 1:
            return float(sr_value[0])
        else:
            raise ValueError
    except ValueError:
        logger.warning(
            f"information found in column {sample_rate_key} "
            "could not be read automatically. Mentioned column contents are:"
            f"\n{sr_value}\n"
            "Sample rate was not set."
        )
        return None


def _detect_sample_rate_from_time_column(data: np.ndarray) -> float | None:
    for i in range(data.shape[0]):
        if _is_likely_time_column(data[i]):
            time_step = float(data[i, 1] - data[i, 0])
            sample_rate = float(1 / time_step)
            return sample_rate
    return None


def _is_likely_eeg_column(column: np.ndarray) -> bool:
    return (not _column_is_monotonic(column)) and (not _column_is_sparse(column))


def _is_likely_time_column(column: np.ndarray) -> bool:
    u_diffs = np.unique(np.diff(column).round(6))
    if len(u_diffs) == 1 and u_diffs[0] > 0:
        return True
    return False


def _leave_only_eeg_channels(data: np.ndarray, ch_names: list[str]) -> tuple[np.ndarray, list[str]]:
    eeg_ix = [i for i in range(data.shape[0]) if _is_likely_eeg_column(data[i])]
    new_channel_names = [x for i, x in enumerate(ch_names) if i in eeg_ix]
    return data[eeg_ix], new_channel_names


def _column_has_eeg_name(column_name: str) -> bool:
    # TODO: #13 should not be case-sensitive?
    return column_name in EEG_MONTAGES['10-10']


def _column_is_monotonic(column: np.ndarray) -> bool:
    diffs = np.diff(column)
    return bool(np.all(diffs >= 0))


def _column_is_sparse(column: np.ndarray, sparsity_threshold: float = 0.2) -> bool:
    u_values, u_counts = np.unique(column, return_counts=True)
    most_common_value_rate = np.max(u_counts) / len(column)
    return bool(most_common_value_rate > sparsity_threshold)


def _find_keys_in_data(keys: list[str], searching_for: list[str]) -> list:
    found = []
    translation = str.maketrans('', '', '_- .')
    to_check = [k.lower().translate(translation) for k in searching_for]
    for k in keys:
        if k.lower().translate(translation) in to_check:
            found.append(k)
    return found



from pathlib import Path
from breesy.load import load_recording
from breesy.recording import select_channels, cut_by_second_range

REPO_ROOT = Path(__file__).parent.parent


def test_recording_select_channels():
    # Arrange
    datapath = REPO_ROOT / "tests" / "data" / "alpha" / "subject_01.mat"
    rec = load_recording(str(datapath))

    # Act
    selected = select_channels(rec, ['Channel_2', 'Channel_3', 'Channel_4'])

    # Assert
    assert selected.data.shape[0] == 3
    assert selected.channel_names == ['Channel_2', 'Channel_3', 'Channel_4']
    assert selected.data.shape[1] == rec.data.shape[1]
    assert selected.sample_rate == rec.sample_rate


def test_recording_cut_by_seconds():
    # Arrange
    datapath = REPO_ROOT / "tests" / "data" / "alpha" / "subject_01.mat"
    rec = load_recording(str(datapath))

    # Act
    cut = cut_by_second_range(rec, start_second=0, end_second=10)

    # Assert
    expected_samples = int(10 * rec.sample_rate) + 1
    assert cut.data.shape == (rec.data.shape[0], expected_samples)
    assert cut.sample_rate == rec.sample_rate
    assert cut.channel_names == rec.channel_names


def test_recording_operations_preserve_shape():
    # Arrange
    datapath = REPO_ROOT / "tests" / "data" / "alpha" / "subject_01.mat"
    rec = load_recording(str(datapath))

    # Act
    processed = select_channels(rec, ['Channel_2', 'Channel_5', 'Channel_10'])
    processed = cut_by_second_range(processed, start_second=5, end_second=15)

    # Assert
    expected_samples = int(10 * rec.sample_rate) + 1
    assert processed.data.shape == (3, expected_samples)
    assert processed.sample_rate == rec.sample_rate
    assert len(processed.channel_names) == 3



"""Tests for loading EEG data in various formats."""

from pathlib import Path
import pytest
from breesy.load import load_recording
from breesy.recording import Recording


def test_load_mat_alpha_dataset():
    """Test loading MAT format using existing alpha dataset.

    This test uses the same subject_01.mat file that students/users can download_example_data("alpha").
    """
    repo_root = Path(__file__).parent.parent
    datapath = repo_root / "tests" / "data" / "alpha" / "subject_01.mat"

    assert datapath.exists(), f"Test data file not found: {datapath}"

    rec = load_recording(str(datapath))

    assert rec is not None, "Recording should not be None"
    assert isinstance(rec, Recording), "Should return a Recording instance"
    assert rec.data is not None, "Recording should have data"
    assert rec.data.ndim == 2, "Data should be 2D (channels x samples)"
    assert rec.sample_rate > 0, "Sample rate should be positive"
    assert len(rec.channel_names) > 0, "Should have channel names"
    assert len(rec.channel_names) == rec.data.shape[0], \
        "Number of channel names should match number of data channels"

    # Format-specific assertions
    assert rec.data.shape == (16, 119808), "Alpha dataset should have 16 channels and 119808 samples"
    assert rec.sample_rate == 512.0, "Alpha dataset sample rate should be 512 Hz"


def test_load_brainvision_ssvep_dataset():
    """Test loading BrainVision format using SSVEP example dataset."""
    repo_root = Path(__file__).parent.parent
    datapath = repo_root / "tests" / "data" / "ssvep" / "sub-01_ses-01_task-ssvep_eeg.vhdr"

    assert datapath.exists(), f"Test data file not found: {datapath}"

    # Verify companion files exist
    eeg_file = datapath.with_suffix('.eeg')
    vmrk_file = datapath.with_suffix('.vmrk')
    assert eeg_file.exists(), f"BrainVision .eeg file not found: {eeg_file}"
    assert vmrk_file.exists(), f"BrainVision .vmrk file not found: {vmrk_file}"

    rec = load_recording(str(datapath))

    assert rec is not None, "Recording should not be None"
    assert isinstance(rec, Recording), "Should return a Recording instance"
    assert rec.data is not None, "Recording should have data"
    assert rec.data.ndim == 2, "Data should be 2D (channels x samples)"
    assert rec.sample_rate > 0, "Sample rate should be positive"
    assert len(rec.channel_names) > 0, "Should have channel names"
    assert len(rec.channel_names) == rec.data.shape[0], \
        "Number of channel names should match number of data channels"

    # Format-specific assertions
    assert rec.data.shape == (32, 469680), "SSVEP dataset should have 32 channels and 469680 samples"
    assert rec.sample_rate == 1000, "SSVEP dataset sample rate should be 1000 Hz"
    assert len(rec.channel_names) == 32, "SSVEP dataset should have 32 channel names"


# Future tests for formats not yet supported
@pytest.mark.skip(reason="EDF format not yet supported by Breesy")
def test_load_edf_schalk_dataset():
    """Test loading EDF format - currently not supported. This is the same schalk dataset that users can download_example_data("schalk").
    """
    repo_root = Path(__file__).parent.parent
    datapath = repo_root / "tests" / "data" / "schalk" / "S001R01.edf"

    # When EDF support is added, uncomment:
    # rec = load_recording(str(datapath))
    # assert rec is not None
    # assert rec.data.shape[0] == 64  # 64 channels
    # assert rec.sample_rate == 160  # 160 Hz


@pytest.mark.skip(reason="EEGLAB .set format not yet supported by Breesy")
def test_load_eeglab_n170_dataset():
    """Test loading EEGLAB format - currently not supported."""
    repo_root = Path(__file__).parent.parent
    datapath = repo_root / "tests" / "data" / "n170" / "1_N170_shifted_ds_reref_ucbip_hpfilt_ica_weighted.set"

    # When EEGLAB support is added, uncomment:
    # rec = load_recording(str(datapath))
    # assert rec is not None



from pathlib import Path
from breesy.load import load_recording
from breesy.recording import select_channels
from breesy.plots import (
    plot_recording,
    plot_mean_frequency_spectrum,
    plot_recording_montage,
    plot_mean_topography,
)

REPO_ROOT = Path(__file__).parent.parent
TEST_OUTPUT_DIR = Path(__file__).parent / "test_outputs"

STANDARD_10_10_CHANNELS = [
    'Fp1', 'Fp2', 'F7', 'F3', 'Fz', 'F4', 'F8', 'FC5', 'FC1', 'FC2',
    'FC6', 'T7', 'C3', 'Cz', 'C4', 'T8', 'CP5', 'CP1', 'CP2', 'CP6',
    'P7', 'P3', 'Pz', 'P4', 'P8', 'O1', 'Oz', 'O2'
]


def test_plot_recording():
    # Arrange
    datapath = REPO_ROOT / "tests" / "data" / "alpha" / "subject_01.mat"
    output_path = TEST_OUTPUT_DIR / "plot_recording.png"
    rec = load_recording(str(datapath))

    # Act
    plot_recording(rec, start=0, duration=10, savepath=str(output_path), save_format="png")

    # Assert
    assert output_path.exists()
    assert output_path.stat().st_size > 0


def test_plot_mean_frequency_spectrum():
    # Arrange
    datapath = REPO_ROOT / "tests" / "data" / "alpha" / "subject_01.mat"
    output_path = TEST_OUTPUT_DIR / "plot_spectrum.png"
    rec = load_recording(str(datapath))

    # Act
    plot_mean_frequency_spectrum(rec, savepath=str(output_path), save_format="png")

    # Assert
    assert output_path.exists()
    assert output_path.stat().st_size > 0


def test_plot_recording_montage():
    # Arrange
    datapath = REPO_ROOT / "tests" / "data" / "ssvep" / "sub-01_ses-01_task-ssvep_eeg.vhdr"
    output_path = TEST_OUTPUT_DIR / "plot_montage.png"
    rec = load_recording(str(datapath))

    # Act
    plot_recording_montage(rec, savepath=str(output_path), save_format="png")

    # Assert
    assert output_path.exists()
    assert output_path.stat().st_size > 0


def test_plot_mean_topography():
    # Arrange
    datapath = REPO_ROOT / "tests" / "data" / "ssvep" / "sub-01_ses-01_task-ssvep_eeg.vhdr"
    output_path = TEST_OUTPUT_DIR / "plot_topography.png"
    rec = load_recording(str(datapath))
    rec = select_channels(rec, STANDARD_10_10_CHANNELS)

    # Act
    plot_mean_topography(rec, start=0, duration=1, savepath=str(output_path), save_format="png")

    # Assert
    assert output_path.exists()
    assert output_path.stat().st_size > 0



import numpy as np

from breesy import breesy_scipy, constants
from breesy.recording import Recording
from breesy.breesy_scipy import _si_butter_bandpass_sos, _si_sosfiltfilt
from breesy.errors import BreesyError
from breesy.recording import update_data
from breesy.type_hints import enforce_type_hints


@enforce_type_hints
def notch_filter(recording: Recording, frequency: int | float | list[int | float], method: str = 'fir',
                 quality_factor: int | float = constants.POWERLINE_NOTCH_QUALITY_FACTOR) -> Recording:
    """Apply notch filter to remove a specific frequency from a recording.

    :param recording: Input recording
    :param frequency: Frequency to remove in Hz
    :param method: Filter method to use. Default is FIR ("fir"). Other option is IIR ("iir") which is faster but less suitable for EEG.
    :param quality_factor: Quality factor for the notch filter

    :return: Filtered recording
    """
    if isinstance(frequency, (int, float)):
       frequency = [frequency]

    new_data = recording.data.copy()

    if method == 'fir':
        new_data = _multi_notch_fir(data=new_data, frequencies=frequency, sample_rate=recording.sample_rate,
                                    notch_width=1.0, transition_width=0.5)  # TODO: connect these with quality factor used in iir?
    elif method == 'iir':
        new_data = _multi_notch_iir(data=new_data, frequencies=frequency, sample_rate=recording.sample_rate,
                                    quality_factor=quality_factor)
    else:
        raise BreesyError(f'The filter named "{method}" is unknown or unsupported in Breesy.',
                            'Use one of supported filters: "fir" (recommended, default) or "iir" (faster).')

    return update_data(recording, new_data)


@enforce_type_hints
def select_bandwidth(recording: Recording, low: float | int, high: float | int, filter_order: int = 5) -> Recording:
    """Filter recording to keep only a specific frequency band.

    :param recording: Input recording
    :param low: Lower frequency bound in Hz
    :param high: Upper frequency bound in Hz
    :param filter_order: Order of the filter

    :return: Filtered recording
    """

    # TODO: update function to FIR and unhardcode

    data = recording.data
    sample_rate = recording.sample_rate

    low_high_range_valid = low <= high
    if not low_high_range_valid:
        raise BreesyError("Low frequency bound is higher than the high frequency bound",
                          "Provide a low frequency bound value that is less than or equal to the high frequency bound")

    high_bound_exceeds_nyquist = high >= (sample_rate / 2)
    if high_bound_exceeds_nyquist:
        raise BreesyError("High frequency bound exceeds Nyquist frequency",
                          "Provide a high frequency bound value that is less than half of the sample rate")

    bandpass = _si_butter_bandpass_sos(low, high, sample_rate, filter_order)

    new_eeg_data = _si_sosfiltfilt(bandpass, data)

    return update_data(recording, new_eeg_data)


def _multi_notch_fir(data: np.ndarray, frequencies: list[float | int], sample_rate: int | float,
                        notch_width: float, transition_width: float) -> np.ndarray:
    bands = [x for freq in frequencies for x in (freq-notch_width/2, freq+notch_width/2)]
    numtaps = int(sample_rate / transition_width)
    numtaps = numtaps if numtaps % 2 == 1 else numtaps + 1
    fir_coeff = breesy_scipy._si_firwin_bandstop_multibands(numtaps=numtaps, bands=bands, sample_rate=sample_rate)
    filtered_data = breesy_scipy._si_filtfilt(fir_coeff, [1], data)
    return filtered_data


def _multi_notch_iir(data: np.ndarray, frequencies: list[float | int], sample_rate: int | float,
                        quality_factor: int | float) -> np.ndarray:
    for freq in frequencies:
        b, a = breesy_scipy._si_iirnotch(freq=freq, quality_factor=quality_factor, sample_rate=sample_rate)
        data = breesy_scipy._si_filtfilt(b=b, a=a, data=data)
    return data



"""Tests for executing tutorial Jupyter notebooks."""

import subprocess
from pathlib import Path
import pytest


def test_tutorial_notebooks_execute():
    """Execute all tutorial notebooks and save HTML outputs."""
    repo_root = Path(__file__).parent.parent
    tutorials_dir = repo_root / "docs" / "breesy-book" / "tutorials"
    output_dir = Path(__file__).parent / "test_outputs"

    notebooks = list(tutorials_dir.glob("*.ipynb"))

    assert len(notebooks) > 0, f"No notebooks found in {tutorials_dir}"

    print(f"\nFound {len(notebooks)} tutorial notebooks to execute:")
    for nb in notebooks:
        print(f"  - {nb.name}")

    # Execute each notebook
    failed_notebooks = []
    for nb_path in notebooks:
        print(f"\nExecuting: {nb_path.name}")

        # --allow-errors allows notebooks with validation issues to execute
        # --ExecutePreprocessor.timeout increases timeout for slow cells
        result = subprocess.run([
            "jupyter", "nbconvert",
            "--to", "html",
            "--execute",
            "--allow-errors",
            "--ExecutePreprocessor.timeout=300",
            "--output-dir", str(output_dir),
            str(nb_path)
        ], capture_output=True, text=True, timeout=600)

        if result.returncode != 0:
            failed_notebooks.append({
                'notebook': nb_path.name,
                'stderr': result.stderr,
                'stdout': result.stdout
            })
            print(f"  ❌ FAILED: {nb_path.name}")
        else:
            print(f"  ✅ SUCCESS: {nb_path.name}")

    # Assert all notebooks executed successfully
    if failed_notebooks:
        error_msg = "\n\nNotebook execution failures:\n"
        for failure in failed_notebooks:
            error_msg += f"\n{failure['notebook']}:\n"
            error_msg += f"STDERR: {failure['stderr']}\n"
            error_msg += f"STDOUT: {failure['stdout']}\n"
        pytest.fail(error_msg)

    print(f"\n✅ All {len(notebooks)} notebooks executed successfully")
    print(f"HTML outputs saved to: {output_dir}")



from breesy.load import load_recording
from breesy.plots import plot_recording
from breesy.errors import BreesyError

raise BreesyError(
    "No info found about sample rate.",
    "Please find the recording's sample rate in data description and set manually by "
    "passing an additional parameter 'sample_rate' into the data loading function. "
    "Example: load_recording(path, sample_rate=128)."
)



"""Breesy - Brainwave Retrieval, Extraction, and Easy Signal Yield.

Breesy provides easy-to-use functions and pipelines for neuroscience students
who want to apply signal processing and statistical methods on EEG data.
"""

__version__ = '0.3.5'

# Import core classes
from .recording import Recording
from .Event import Event
from .RecordingMetadata import RecordingMetadata

# Make key modules and functions available at package level
from . import constants, events, features, errors, filtering, \
    load, load_events, playground, plots, processing, recording, type_hints, log



import numpy as np
from scipy import signal as si
from scipy.interpolate import RBFInterpolator

from .errors import protect_from_lib_error
from . import constants


@protect_from_lib_error("scipy")
def _si_welch(data: np.ndarray, sample_rate: int | float, window: str | None = None,
              scaling: str = 'density', average: str = 'mean') -> tuple[np.ndarray, np.ndarray]:
    window = window or 'hann'
    return si.welch(x=data, fs=sample_rate, window=window, scaling=scaling, average=average, axis=-1)


@protect_from_lib_error("scipy")
def _si_periodogram(data: np.ndarray, sample_rate: int | float, window: str | None = None,
                    scaling: str = 'density') -> tuple[np.ndarray, np.ndarray]:
    window = window or 'boxcar'
    return si.periodogram(x=data, fs=sample_rate, window=window, scaling=scaling, axis=-1)


@protect_from_lib_error("scipy")
def _si_filtfilt(b, a, data: np.ndarray) -> np.ndarray:
    return si.filtfilt(b=b, a=a, x=data, axis=-1)


@protect_from_lib_error("scipy")
def _si_sosfiltfilt(sos: np.ndarray, data: np.ndarray) -> np.ndarray:
    return si.sosfiltfilt(sos=sos, x=data)


@protect_from_lib_error("scipy")
def _si_butter_bandpass_sos(low: float, high: float, sample_rate: int | float,
                            filter_order: int = constants.BUTTER_FILTER_ORDER) -> np.ndarray:
    return si.butter(N=filter_order, Wn=(low, high), btype='bandpass', output='sos', fs=sample_rate, analog=False)
    # with "sos", always returns single np.ndarray


@protect_from_lib_error("scipy")
def _si_butter_lowpass_sos(freq: float, sample_rate: int | float,
                           filter_order: int = constants.BUTTER_FILTER_ORDER) -> np.ndarray:
    return si.butter(N=filter_order, Wn=freq, btype='lowpass', output='sos', fs=sample_rate, analog=False)
    # with "sos", always returns single np.ndarray


@protect_from_lib_error("scipy")
def _si_butter_highpass_sos(freq: float, sample_rate: int | float,
                            filter_order: int = constants.BUTTER_FILTER_ORDER) -> np.ndarray:
    return si.butter(N=filter_order, Wn=freq, btype='highpass', output='sos', fs=sample_rate, analog=False)
    # with "sos", always returns single np.ndarray


@protect_from_lib_error("scipy")
def _si_firwin_bandpass_multibands(numtaps: int, bands: list[int | float], sample_rate: int | float, window: str | None = None):
    window = window or "hamming"
    return si.firwin(numtaps=numtaps, cutoff=bands, window=window, pass_zero=False, fs=sample_rate)


@protect_from_lib_error("scipy")
def _si_firwin_bandstop_multibands(numtaps: int, bands: list[int | float], sample_rate: int | float, window: str | None = None):
    window = window or "hamming"
    return si.firwin(numtaps=numtaps, cutoff=bands, window=window, pass_zero='bandstop', fs=sample_rate)


@protect_from_lib_error("scipy")
def _si_iirnotch(freq: float, quality_factor: float, sample_rate: int | float) -> tuple[np.ndarray, np.ndarray]:
    return si.iirnotch(w0=freq, Q=quality_factor, fs=sample_rate)


@protect_from_lib_error("scipy")
def _si_get_rbf_interpolator(coords_arr: np.ndarray, values_arr: np.ndarray) -> RBFInterpolator:
    return RBFInterpolator(coords_arr, values_arr, kernel='thin_plate_spline')


import numpy as np

from .Event import Event
from .RecordingMetadata import RecordingMetadata
from .errors import BreesyError
from .type_hints import enforce_type_hints
from .log import logger


class Recording:
    """Represents an EEG recording with data, channel information, and events."""

    data: np.ndarray
    """EEG data array of shape (channels, samples)"""

    channel_names: list[str]
    """List of channel names"""

    sample_rate: int | float
    """Sampling rate in Hz"""

    events: list[Event] | None
    """List of Event objects"""

    metadata: RecordingMetadata | None
    """Additional metadata about the recording"""

    @enforce_type_hints
    def __init__(self,
                 data: np.ndarray,
                 channel_names: list[str],
                 sample_rate: int | float,
                 events: list[Event] | None = None,
                 metadata: RecordingMetadata | None = None, ):
        """Initialize a Recording.

        :param data: EEG data array of shape (channels, samples)
        :param channel_names: List of channel names
        :param sample_rate: Sampling rate in Hz
        :param events: List of Event objects
        :param metadata: Additional metadata about the recording
        """

        if data.ndim != 2:
            raise BreesyError(
                "The EEG data provided has an incorrect number of dimensions. Expected data.ndim to be 2, but got "
                f"{data.ndim} dimensions instead.",
                "The data provided is likely not EEG data, or there are several records joined into one file. However, Breesy only supports ordinary two-dimensional EEG data for now."
            )

        n_channels, n_samples = data.shape
        if (n_channels != len(channel_names)) and (n_samples != len(channel_names)):
            raise BreesyError(
                f"The number of channel names ({len(channel_names)}) provided does not match the "
                f"number of channels ({n_channels}) in the EEG data.",
                "Make sure that the number of channel names provided matches the number of channels in the EEG data."
                "You can see the number of channel names like this: len(eeg_channel_names)."
            )
        elif (n_channels != len(channel_names)) and (n_samples == len(channel_names)):
            logger.warning('data rows and columns will be switched places according to the provided channel names.')
            self.data = data.T
        else:
            self.data = data

        self.channel_names = channel_names

        if sample_rate <= 0:
            raise BreesyError(
                f"Sample rate should be more than 0, but you provided {sample_rate}.",
                "Make sure that the sample rate is a positive integer. E.g., a sample rate of: 512."
            )

        is_typical_sample_rate = 100 < sample_rate < 5000
        if not is_typical_sample_rate:
            logger.warning(f"Unusual sample rate ({sample_rate} Hz). Typical EEG sample rates range from 128-2000 Hz.")
        self.sample_rate = sample_rate

        if events is None:
            events = []
        self._set_events(events)

        self.metadata = metadata

    def _set_events(self, events: list[Event]) -> None:
        out_of_range_events = [event for event in events if event.index >= self.number_of_samples]
        if out_of_range_events:
            raise BreesyError(
                f"Found {len(out_of_range_events)} events with indices out of range for the provided EEG data. "
                f"First Got index {out_of_range_events[0].index} while the EEG data has {self.number_of_samples} samples.",
                "Make sure that the event indices are within the range of the EEG data. "
                "For instance, if the EEG data has 10000 samples, the event indices should be less than 10000. "
                "You can see the number of samples in the EEG data like so: eeg_data.shape[1]."
            )

        sorted_events = sorted(events, key=lambda event: event.index)
        self.events = sorted_events

    def __repr__(self):
        """Return a string representation of the Recording object."""
        metadata_repr = f", name={self.metadata.name}" if self.metadata else ""
        events_repr = f", event_count={len(self.events)}" if self.events else ""
        return (f"Recording("
                f"channels={self.number_of_channels}, "
                f"duration={self.duration:.2f} s, "
                f"sample_rate={self.sample_rate} Hz"
                f"{metadata_repr}"
                f"{events_repr}"
                f")")


    @property
    def number_of_channels(self):
        """Return the number of channels in the recording."""
        return self.data.shape[0]

    @property
    def duration(self):
        """Return the duration of the recording in seconds."""
        return self.data.shape[1] / self.sample_rate

    @property
    def number_of_samples(self):
        """Return the number of samples in the recording."""
        return self.data.shape[1]


@enforce_type_hints
def cut_by_sample_range(recording: Recording, start_index: int, end_index: int) -> Recording:
    """Cut a recording to a specific sample range.

    :param recording: Input recording
    :param start_index: Start sample index
    :param end_index: End sample index

    :return: New recording containing only the specified sample range
    """
    if start_index < 0 or end_index > recording.number_of_samples:
        raise BreesyError(
            f"Requested to cut from {start_index} to {end_index}, but the recording has only {recording.number_of_samples} samples.",
            "Make sure that the start and end sample indices are within the range of the recording samples. "
            "You can see the number of samples in the recording like so: recording.number_of_samples."
        )

    # Take both ends inclusive so that the events themselves also fit in the epoch
    new_eeg_data = recording.data[:, start_index:end_index + 1]
    events_in_new_data = [event for event in (recording.events or []) if start_index <= event.index <= end_index]
    events_with_offset = [Event(e.name, e.index - start_index) for e in events_in_new_data]

    return Recording(
        data=new_eeg_data,
        channel_names=recording.channel_names,
        sample_rate=recording.sample_rate,
        events=events_with_offset,
        metadata=recording.metadata
    )


@enforce_type_hints
def cut_by_second_range(recording: Recording, start_second: float | int, end_second: float | int):
    """Cut a recording to a specific time range.

    :param recording: Input recording
    :param start_second: Start time in seconds
    :param end_second: End time in seconds

    :return: New recording containing only the specified time range
    """
    if start_second > end_second:
        raise BreesyError(
            f"Requested to cut from {start_second} s to {end_second} s, but the start time is higher than the end time.",
            "Make sure that the start time is lower than the end time."
        )

    if end_second > recording.duration:
        raise BreesyError(
            f"Requested to cut from {start_second} s to {end_second} s, but the recording has only {recording.duration}s.",
            "Make sure that the start and end seconds are within the range of the recording duration. "
            "You can see the recording duration like so: recording.duration."
        )

    start_index = int(recording.sample_rate * start_second)
    end_index = int(recording.sample_rate * end_second)

    return cut_by_sample_range(recording, start_index, end_index)


@enforce_type_hints
def add_events(recording: Recording, events: list[Event]) -> Recording:
    """Create a new recording with additional events.

    :param recording: Input recording
    :param events: List of events to add

    :return: New recording with added events
    """
    new_events = recording.events or [] + events
    return Recording(
        data=recording.data,
        channel_names=recording.channel_names,
        sample_rate=recording.sample_rate,
        events=new_events,
        metadata=recording.metadata
    )


@enforce_type_hints
def remove_channels(recording: Recording, channels_to_remove: list[str]) -> Recording:
    """Create a new recording with specified channels removed.

    :param recording: Input recording
    :param channels_to_remove: List of channel names to remove

    :return: New recording without the specified channels
    """
    missing_channels = [channel for channel in channels_to_remove if channel not in recording.channel_names]
    if missing_channels:
        raise BreesyError(
            f"Channels to remove contain channel names that are not present in the recording, namely: {missing_channels}.",
            "Make sure that the channel names provided to remove are present in the recording. "
            "Double check that you haven't removed those channels for this data already. "
            "You can see the channel names in the recording with: recording.channel_names."
        )

    channel_indices_to_remove = [recording.channel_names.index(channel) for channel in channels_to_remove]
    new_eeg_data = np.delete(recording.data, channel_indices_to_remove, axis=0)
    new_channel_names = [channel for channel in recording.channel_names if channel not in channels_to_remove]
    return Recording(
        data=new_eeg_data,
        channel_names=new_channel_names,
        sample_rate=recording.sample_rate,
        events=recording.events,
        metadata=recording.metadata
    )


@enforce_type_hints
def select_channels(recording: Recording, channels_to_select: list[str]) -> Recording:
    """Create a new recording with only the specified channels.

    :param recording: Input recording
    :param channels_to_select: List of channel names to keep

    :return: New recording with only the specified channels
    """
    missing_channels = [channel for channel in channels_to_select if channel not in recording.channel_names]
    if missing_channels:
        raise BreesyError(
            f"Channels to select contain channel names that are not present in the recording, namely: {missing_channels}.",
            "Make sure that the channel names provided to select are present in the recording. "
            "You can see the channel names in the recording with: recording.channel_names."
        )

    channel_indices_to_select = [recording.channel_names.index(channel) for channel in channels_to_select]
    new_eeg_data = recording.data[channel_indices_to_select]
    return Recording(
        data=new_eeg_data,
        channel_names=channels_to_select,
        sample_rate=recording.sample_rate,
        events=recording.events,
        metadata=recording.metadata
    )


@enforce_type_hints
def set_channel_names(recording: Recording, ch_names: list[str]) -> Recording:
    """Set new channel names for a recording.

    :param recording: Input recording
    :param ch_names: List of new channel names

    :return: New recording with updated channel names
    """
    if len(ch_names) != recording.number_of_channels:
        raise BreesyError(
            f"{len(ch_names)} channel names are provided, but data has {recording.number_of_channels} channels. "
            "Please recheck your list of channel names. "
        )
    return Recording(
        data=recording.data,
        channel_names=ch_names,
        sample_rate=recording.sample_rate,
        events=recording.events,
        metadata=recording.metadata
    )


@enforce_type_hints
def get_events_by_name(recording: Recording, event_name: str) -> list[Event]:
    """Return a list of events with a specific name.

    :param recording: Input recording
    :param event_name: Name of events to find

    :return: List of events with the specified name
    """
    return [event for event in recording.events if event.name == event_name]


@enforce_type_hints
def split_by_continuous_events(recording: Recording, start_event_name: str, end_event_name: str) -> list[Recording]:
    """Split a recording into segments between start and end events.

    :param recording: Input recording
    :param start_event_name: Name of events marking segment starts
    :param end_event_name: Name of events marking segment ends

    :return: List of recording segments between each start-end event pair
    """
    if not recording.events:
        raise BreesyError(
            "No events found in the recording.",
            "Make sure that the recording has events. You can see the number of events like so: len(recording.events)."
        )

    start_events = get_events_by_name(recording, start_event_name)
    end_events = get_events_by_name(recording, end_event_name)

    if len(start_events) != len(end_events):
        raise BreesyError(
            f"Found {len(start_events)} start events and {len(end_events)} end events. "
            "Expected the number of start and end events to be the same.",
            "Make sure each start event has a corresponding end event. "
            "You can see the number of start events like so: get_events_by_name(recording, start_event_name); "
            "number of end events: get_events_by_name(recording, end_event_name)."
        )

    epoch_recordings = []
    for start_event, end_event in zip(start_events, end_events):
        if start_event.index >= end_event.index:
            pair_number = start_events.index(start_event) + 1
            raise BreesyError(
                f"When going pairwise through events, found a pair (nr. {pair_number}) where start event index ({start_event.index}) is higher than the end event index ({end_event.index}).",
                "Inspect the events manually, for start events: get_events_by_name(recording, start_event_name); "
                "end events: get_events_by_name(recording, end_event_name). "
                "Make sure that the start event index is lower than the end event index. "
            )

        epoch_recording = cut_by_sample_range(recording, start_event.index, end_event.index)
        epoch_recordings.append(epoch_recording)

    return epoch_recordings


@enforce_type_hints
def split_by_events(recording: Recording, event_name: str, duration: float | int = 5.0) -> list[Recording]:
    """Split a recording into segments starting at specified events.

    :param recording: Input recording
    :param event_name: Name of events to use as segment starts
    :param duration: Duration of each segment in seconds

    :return: List of recording segments starting at each event
    """
    events = get_events_by_name(recording, event_name)

    epoch_recordings = []
    for event in events:
        end_index = int(event.index + recording.sample_rate * duration)
        epoch_recording = cut_by_sample_range(recording, event.index, end_index)
        epoch_recordings.append(epoch_recording)

    return epoch_recordings


@enforce_type_hints
def split_by_window_duration(recording: Recording, window_duration: float | int, overlap_duration: float | int = 0.0) -> list[Recording]:
    """Split a recording into equal-sized segments of specified duration with optional overlap.

    :param recording: Input recording
    :param window_duration: Duration of each segment in seconds
    :param overlap_duration: Duration of overlap between consecutive segments in seconds (default: 0.0)

    :return: List of equal-sized recording segments
    """
    if overlap_duration < 0:
        raise BreesyError(
            f"Overlap duration ({overlap_duration}s) cannot be negative.",
            "Provide a non-negative overlap duration value."
        )

    if overlap_duration >= window_duration:
        raise BreesyError(
            f"Overlap duration ({overlap_duration}s) must be less than window duration ({window_duration}s).",
            "Provide an overlap duration that is smaller than the window duration."
        )

    num_samples_per_epoch = int(window_duration * recording.sample_rate)

    if num_samples_per_epoch > recording.number_of_samples:
        raise BreesyError(
            f"Epoch duration ({window_duration}s) is longer than the recording duration ({recording.duration}s).",
            "Make sure that the epoch duration is shorter than the recording duration. "
            "You can see the recording duration like so: recording.duration."
        )

    step_duration = window_duration - overlap_duration
    step_samples = int(step_duration * recording.sample_rate)

    max_start_sample = recording.number_of_samples - num_samples_per_epoch
    num_epochs = (max_start_sample // step_samples) + 1

    epoch_recordings = []
    for i in range(num_epochs):
        start_index = i * step_samples
        end_index = start_index + num_samples_per_epoch

        # skip window if out of bounds
        if end_index > recording.number_of_samples:
            break

        epoch_data = recording.data[:, start_index:end_index]
        epoch_recording = Recording(
            data=epoch_data,
            channel_names=recording.channel_names,
            sample_rate=recording.sample_rate,
            events=recording.events,
            metadata=recording.metadata
        )
        epoch_recordings.append(epoch_recording)

    return epoch_recordings


@enforce_type_hints
def update_data(recording: Recording, new_eeg_data: np.ndarray) -> Recording:
    """Create a new recording with updated EEG data.

    :param recording: Input recording
    :param new_eeg_data: New EEG data array to use

    :return: New recording with updated data but same metadata and events
    """
    if recording.number_of_samples != new_eeg_data.shape[1]:
        raise BreesyError(
            f"New EEG data has a different number of samples ({new_eeg_data.shape[1]}) than the original EEG data ({recording.number_of_samples}).",
            "Make sure that the new EEG data has the same number of samples as the original EEG data. "
            "You can see the number of samples in the original EEG data like so: recording.number_of_samples."
        )

    return Recording(
        data=new_eeg_data,
        channel_names=recording.channel_names,
        sample_rate=recording.sample_rate,
        events=recording.events,
        metadata=recording.metadata
    )



import numpy as np
from matplotlib import pyplot as plt
from scipy import fft

from . import constants
from .errors import protect_from_lib_error
from .type_hints import enforce_type_hints


WAVE_SAMPLE_RATE = constants.PLAYGROUND_WAVE_SAMPLE_RATE
WAVE_T = np.arange(0, constants.PLAYGROUND_WAVE_DURATION, 1/WAVE_SAMPLE_RATE)
SPECTRUM_F = fft.rfftfreq(len(WAVE_T), d=1/WAVE_SAMPLE_RATE)


class Wave:
  def __init__(self, a):
    self.a = np.round(a, constants.PLAYGROUND_Y_ROUND_FACTOR)

  def __add__(self, wave2):
    return Wave(a=self.a+wave2.a)

  def __sub__(self, wave2):
    return Wave(a=self.a-wave2.a)

  def __mul__(self, x: int | float):
    return Wave(a=self.a*x)

  def __truediv__(self, x: int | float):
    return Wave(a=self.a/x)

  def __eq__(self, wave2) -> bool:
    return np.allclose(self.a, wave2.a, rtol=1e-05, atol=1e-08)
  

class Spectrum:
  def __init__(self, s):
    self.s = s

  def __add__(self, spectrum2):
    return Spectrum(s=self.s+spectrum2.s)

  def __sub__(self, spectrum2):
    return Spectrum(s=self.s-spectrum2.s)

  def __mul__(self, x: int | float):
    return Spectrum(s=self.s*x)

  def __truediv__(self, x: int | float):
    return Spectrum(s=self.s/x)

  def __eq__(self, spectrum2) -> bool:
    return np.allclose(self.s, spectrum2.s, rtol=1e-05, atol=1e-08)


@enforce_type_hints
def create_wave(frequency: int | float = 1,
                amplitude: int | float = 1,
                phase: int | float = 0) -> Wave:
    f"""Create a simple sine wave with provided attributes. The wave will have length of {constants.PLAYGROUND_WAVE_DURATION} second and sample rate of {WAVE_SAMPLE_RATE} Hz.

    :param frequency: Frequency in Hz (Default: 1 Hz)
    :param amplitude: Amplitude (Default: 1 arbitrary unit)
    :param phase: Phase of the wave (Default: 0)
    """
    assert frequency <= WAVE_SAMPLE_RATE / 2, f"Default sample rate is {WAVE_SAMPLE_RATE} Hz, so the maximum allowed frequency is {WAVE_SAMPLE_RATE/2} Hz."
    a = amplitude * np.sin(2 * np.pi * frequency * WAVE_T + phase)
    return Wave(a=a)


@enforce_type_hints
@protect_from_lib_error("numpy")
def generate_white_noise(random_state: int | None = None, scale: int | float = 1) -> Wave:
    """Generate Gaussian white noise.

    :param random_state: Random seed for reproducibility
    :param scale: standard deviation of data (Default: 1 arbitrary unit)
    """
    rng = np.random.default_rng(random_state)
    n_samples = len(WAVE_T)
    noise = rng.normal(0, scale, n_samples)
    return Wave(a=noise)


@enforce_type_hints
@protect_from_lib_error("numpy")
def generate_pink_noise(random_state: int | None = None, scale: int | float = 1) -> Wave:
    """Generate pink (1/f) noise.

    :param random_state: Random seed for reproducibility
    :param scale: standard deviation of data (Default: 1 arbitrary unit)
    """
    rng = np.random.default_rng(random_state)
    n_samples = len(WAVE_T)

    white_noise = rng.normal(0, 1, n_samples)
    f = np.fft.fftfreq(n_samples)
    f[0] = np.inf  # Avoid division by zero
    f = np.abs(f)
    pink_noise = np.real(np.fft.ifft(np.fft.fft(white_noise) / np.sqrt(f)))
    pink_noise = pink_noise / np.std(pink_noise) * scale
    # TODO: does scale even do anything here?
    return Wave(a=pink_noise)


@enforce_type_hints
def wave_to_spectrum(wave: Wave) -> Spectrum:
    """Get a spectrum from a sine wave.

    :param wave: a Wave object
    """
    fft_complex = fft.rfft(wave.a)
    return Spectrum(s=fft_complex)


@enforce_type_hints
def spectrum_to_wave(spectrum: Spectrum) -> Wave:
    """Get a sine wave from a spectrum.

    :param spectrum: a Spectrum object
    """
    a = fft.irfft(spectrum.s)
    return Wave(a=a)


@enforce_type_hints
def plot_waves(*args: Wave) -> None:
    """Plot one or more sine waves overlapping in the same plot.

    :param args: one or more Wave objects
    """
    plt.figure(figsize=(12, 4))
    for i, wave in enumerate(args):
        plt.plot(WAVE_T, wave.a, alpha=0.75, lw=2, label=f'wave {i+1}')
    plt.xlabel('Time (s)')
    plt.ylabel('Amplitude')
    plt.xlim(0, constants.PLAYGROUND_WAVE_DURATION)
    plt.legend(bbox_to_anchor=(1, 1))
    plt.grid(alpha=0.6, ls=':')
    plt.box(False)
    plt.tight_layout()
    plt.show()


@enforce_type_hints
def plot_spectrum(spectrum: Spectrum) -> None:
    """Plot a spectrum.

    :param spectrum: one Spectrum object
    """
    plt.figure(figsize=(12, 4))
    power = np.abs(spectrum.s)**2 / len(WAVE_T)
    last_nonzero = max(np.nonzero(power.round(6) > 0)[0])
    if last_nonzero < 100:
        plt.plot(SPECTRUM_F, power, alpha=0.5, lw=1, ls=':', color=constants.COLOR_SPECTRUM)
        plt.scatter(SPECTRUM_F, power, alpha=0.75, lw=2, s=10, color=constants.COLOR_SPECTRUM)
    else:
        plt.plot(SPECTRUM_F, power, alpha=0.75, lw=2, color=constants.COLOR_SPECTRUM)
    plt.xlabel('Frequency (Hz)')
    plt.ylabel('Amplitude')

    if last_nonzero + 1 < len(SPECTRUM_F):
      xlim_max = min(SPECTRUM_F[last_nonzero+1] * 1.5, SPECTRUM_F.max())
    else:
       xlim_max = min(SPECTRUM_F[-1] * 1.5, SPECTRUM_F.max())
    plt.xlim(0, xlim_max)
    plt.grid(alpha=0.6, ls=':')
    plt.box(False)
    plt.tight_layout()
    plt.show()



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.axes import Axes  # for typing
from tqdm.auto import tqdm

from . import constants
from .recording import Recording, select_channels, cut_by_sample_range, cut_by_second_range
from .processing import get_frequency_spectrum, get_ica_components
from .filtering import select_bandwidth
from .breesy_scipy import _si_get_rbf_interpolator
from .errors import BreesyError
from .type_hints import enforce_type_hints
from .case_insensitive_dict import MutableMapping  # for typing
from .log import logger


LENGTH_PLOT_PARAMS = """
:param start: Start time in seconds for the plot
:param duration: Duration in seconds to plot; use -1 to use the whole recording
"""

MONTAGE_PLOT_PARAMS = """
:param head_color: Fill color for the head outline
:param edge_color: Color for head outline
:param fontsize: Font size for electrode labels
"""

MONTAGE_ELECTRODE_COLORS_PARAMS = """
:param electrode_color: Color to fill electrodes
:param electrode_highlight_color: Color to fill highlighted electrodes
"""

COLOR_CYCLE_PARAMS = """
:param cmap: Name of a Matplotlib colormap to use (see https://matplotlib.org/stable/users/explain/colors/colormaps.html for more information)
:param cmap_size: Number of colors to use from the colormap (useful for continuous colormaps)
"""

HEATMAP_PARAMS = """
:param cmap: Name of a Matplotlib colormap to use (see https://matplotlib.org/stable/users/explain/colors/colormaps.html for more information); recommended to use a diverging colormap
:param vmin: Minimum value for color scaling
:param vmax: Maximum value for color scaling
"""

COMMON_PLOT_PARAMS = """
:param savepath: Path to file if you want to save the plot (not saved by default)
:param save_format: Format for saving the plot into a file (default is PDF which is lossless hense publication-ready)
:param figsize: Size of the figure in inches as a tuple: (width, height); each plot has its own recommended figsize used by default
"""

TOPOGRAPHY_PARAMS = """
:param grid_resolution: Resolution of interpolation grid
:param contour_levels: Number of contour levels for visualization
:param contour_alpha: Alpha of the contour lines; set to 0 to remove countours
:param contour_color: Color of contour lines
:param contour_width: Width of contour lines
:param show_electrodes: Whether to show electrode positions
:param electrode_color: Color for electrode markers
:param electrode_size: Size of electrode markers
:param label_electrodes: Whether to label electrodes with channel names
"""

# -------- signal plots


@enforce_type_hints
def plot_recording(recording: Recording,                   
                   start: float | int = 0.0, duration: float | int | None = 60.0,
                   savepath: str | None = None, save_format: str = "pdf",
                   figsize: tuple[int | float, int | float] | None = None,
                   scaling: float | int = 1.0,
                   cmap: str = 'Dark2', cmap_size: int = 8) -> None:
    f"""Plot EEG signals with channels spread vertically.

    :param recording: Input recording
    {LENGTH_PLOT_PARAMS}
    {COMMON_PLOT_PARAMS}
    :param scaling: decrease if signals overlap too much, or increase if signals are not detailed enough
    {COLOR_CYCLE_PARAMS}
    """

    # TODO: add event visualization

    if figsize is None:
        figsize = (12, recording.number_of_channels / 2)
    fig, ax = plt.subplots(figsize=figsize)

    if (duration is None) or (duration > recording.duration):
        _duration = recording.duration
    else:
        _duration = duration

    _plot_signal(ax=ax, data=recording.data, number_of_samples=recording.number_of_samples, sample_rate=recording.sample_rate, channel_names=recording.channel_names, 
                 start=start, duration=_duration, scaling=scaling, cmap=cmap, cmap_size=cmap_size)
    _finalize_plot(savepath=savepath, save_format=save_format)


@enforce_type_hints
def plot_recording_channels(recording: Recording,
                            channel_names: list[str],
                            start: float | int = 0.0, duration: float | int | None = 60.0,
                            savepath: str | None = None, save_format: str = "pdf",
                            figsize: tuple[int | float, int | float] | None = None,
                            scaling: float | int = 1.0, cmap: str = 'Dark2', cmap_size: int = 8) -> None:
    f"""Plot EEG signals for selected channels only.

    :param recording: Input recording
    :param channel_names: List of channel names to include in the plot
    {LENGTH_PLOT_PARAMS}
    {COMMON_PLOT_PARAMS}
    :param scaling: decrease if signals overlap too much, or increase if signals are not detailed enough
    {COLOR_CYCLE_PARAMS}
    """

    if figsize is None:
        figsize = (12, recording.number_of_channels / 2)
    fig, ax = plt.subplots(figsize=figsize)

    subrecording = select_channels(recording, channel_names)

    if (duration is None) or (duration > subrecording.duration):
        _duration = subrecording.duration
    else:
        _duration = duration

    _plot_signal(ax=ax, data=subrecording.data, number_of_samples=subrecording.number_of_samples, sample_rate=subrecording.sample_rate, channel_names=subrecording.channel_names, 
                 start=start, duration=_duration, scaling=scaling, cmap=cmap, cmap_size=cmap_size)
    _finalize_plot(savepath=savepath, save_format=save_format)


@enforce_type_hints
def plot_decomposed_signal(recording: Recording,
                           bandwidths: dict[str, tuple[float | int, float | int]] | None = None,
                           start: float | int = 0, duration: float | int = 3.0,                           
                           savepath: str | None = None, save_format: str = "pdf",
                           figsize: tuple[int | float, int | float] | None = None,
                           cmap: str = 'Dark2', cmap_size: int = 8) -> None:
    f"""Plot recording signal decomposed into frequency bands.

    :param recording: Input recording
    :param bandwidths: Dictionary of band names and frequency ranges (will use classical bandwidths by default, such as alpha, beta etc.)
    {LENGTH_PLOT_PARAMS}
    {COMMON_PLOT_PARAMS}
    {COLOR_CYCLE_PARAMS}
    """

    if bandwidths is None:
        bandwidths = constants.CLASSIC_BANDWIDTHS   

    start_sample, n_samples, time = _get_timepoints(number_of_samples=recording.number_of_samples, sample_rate=recording.sample_rate, start=start, duration=duration)

    if figsize is None:
        figsize = (12, 2 * len(bandwidths))

    fig, axes = plt.subplots(nrows=len(bandwidths), ncols=1, figsize=figsize, sharex=True, sharey=True)
    
    for (name, (low, high)), ax in zip(bandwidths.items(), axes):
        color_cycle = _get_color_cycle(cmap=cmap, cmap_size=cmap_size)
        filtered = select_bandwidth(recording, low, high)
        for i in range(len(recording.channel_names)):
            ax.plot(time, filtered.data[i, start_sample:n_samples+start_sample], lw=1, alpha=0.8, color=next(color_cycle))
        ax.set_title(f'{name} ({low}-{high} Hz)')
        ax.set_xlim(start, duration+start)
        ax.grid(True)

    plt.xlabel('Time (s)')
    _finalize_plot(savepath=savepath, save_format=save_format)


# -------- frequency plots


@enforce_type_hints
def plot_mean_frequency_spectrum(recording: Recording,
                                 start: float | int = 0, duration: float | int = -1,
                                 savepath: str | None = None, save_format: str = "pdf",
                                 figsize: tuple[int | float, int | float] | None = None,
                                 color: str = constants.COLOR_SPECTRUM) -> None:
    f"""Plot mean frequency spectrum calculated from all channels. Method used: periodogram.

    :param recording: Input EEG recording
    {LENGTH_PLOT_PARAMS}
    {COMMON_PLOT_PARAMS}
    :param color: Color for the spectrum line
    """

    if figsize is None:
        figsize = (12, 4)
    fig, ax = plt.subplots(figsize=figsize)
    ax.grid(True)
    plt.box(False)

    if (start != 0) or (duration != -1):
        start_sample, n_samples, _ = _get_timepoints(number_of_samples=recording.number_of_samples, sample_rate=recording.sample_rate, 
                                                        start=start, duration=duration)
        subrecording = cut_by_sample_range(recording, start_sample, start_sample+n_samples)
    else:
        subrecording = recording

    f, periodograms = get_frequency_spectrum(recording)
    mean_periodogram = periodograms.mean(axis=0)
    geom_mean = np.exp(np.mean(np.log(mean_periodogram)))

    ax.semilogy(f, mean_periodogram, lw=1, color=color)
    ax.set_title(f'Mean periodogram of {len(subrecording.channel_names)} EEG channels')
    ax.set_ylim(geom_mean**2 / mean_periodogram.max(), mean_periodogram.max())
    ax.set_xlim(0, subrecording.sample_rate / 2)
    ax.set_xlabel('Frequency (Hz)')
    ax.set_ylabel('Power')
    _finalize_plot(savepath=savepath, save_format=save_format)


@enforce_type_hints
def plot_frequency_spectrum(recording: Recording,
                            start: float | int = 0, duration: float | int = -1,
                            savepath: str | None = None, save_format: str = "pdf",
                            figsize: tuple[int | float, int | float] | None = None,
                            cmap: str = 'Dark2', cmap_size: int = 8) -> None:
    f"""Plot frequency spectrum for all channels in one plot. Method used: periodogram.

    :param recording: Input EEG recording
    {LENGTH_PLOT_PARAMS}
    {COMMON_PLOT_PARAMS}
    {COLOR_CYCLE_PARAMS}
    """

    if figsize is None:
        figsize = (12, 4)
    fig, ax = plt.subplots(figsize=figsize)
    ax.grid(True)
    plt.box(False)

    if (start != 0) or (duration != -1):
        start_sample, n_samples, _ = _get_timepoints(number_of_samples=recording.number_of_samples, sample_rate=recording.sample_rate, 
                                                        start=start, duration=duration)
        subrecording = cut_by_sample_range(recording, start_sample, start_sample+n_samples)
    else:
        subrecording = recording

    f, periodograms = get_frequency_spectrum(subrecording)
    geom_mean = np.mean(np.exp(np.mean(np.log(periodograms), axis=1)))

    nch = len(subrecording.channel_names)
    color_cycle = _get_color_cycle(cmap=cmap, cmap_size=cmap_size)
    for i in range(nch):
        ax.semilogy(f, periodograms[i], lw=1, color=next(color_cycle), alpha=max(0.2, min(2/nch, 0.9)))
    ax.set_title(f'Periodogram ({nch} EEG channels)')
    ax.set_xlim(0, subrecording.sample_rate / 2)
    ax.set_ylim(geom_mean**2 / periodograms.max(), periodograms.max())
    ax.set_xlabel('Frequency (Hz)')
    ax.set_ylabel('Power')
    _finalize_plot(savepath=savepath, save_format=save_format)


# ---- montage plots


@enforce_type_hints
def plot_recording_montage(recording: Recording,
                              savepath: str | None = None, save_format: str = "pdf",
                              figsize: tuple[int | float, int | float] | None = None,
                              hide_unused: bool = True, to_highlight: list[str] | None = None,
                              head_color: str = constants.COLOR_HEAD_FILL,
                              edge_color: str = constants.COLOR_HEAD_EDGE,
                              fontsize: str | int | float = 10,
                              electrode_color: str = constants.COLOR_ELECTRODE,
                              electrode_highlight_color: str = constants.COLOR_HIGHLIGHT_ELECTRODE) -> None:
    f"""Plot electrode montage showing electrode positions for the recording channels.

    :param recording: Input EEG recording
    {COMMON_PLOT_PARAMS}
    :param hide_unused: Whether to hide electrodes not present in the recording
    :param to_highlight: Which electrodes to highlight with different color (none will be highlighted by default)
    {MONTAGE_PLOT_PARAMS}
    {MONTAGE_ELECTRODE_COLORS_PARAMS}
    """

    montage_name = _detect_montage_type(recording.channel_names)
    locations = _get_locations_for_montage(montage_name)

    if figsize is None:
        figsize = (8, 8)
    fig, ax = plt.subplots(figsize=figsize)

    electrode_names = list(locations.keys())
    electrode_colors = _get_electrode_colors(used_names=recording.channel_names, 
                                             to_show=recording.channel_names if hide_unused else electrode_names, 
                                             to_highlight=to_highlight or [],
                                             head_color=head_color, electrode_color=electrode_color, electrode_highlight_color=electrode_highlight_color)
    _plot_electrode_montage(
        ax=ax,
        locations=locations, electrode_colors=electrode_colors,
        head_color=head_color, edge_color=edge_color, fontsize=fontsize
    )
    _finalize_plot(savepath=savepath, save_format=save_format)


@enforce_type_hints
def plot_default_montage(montage: str = "10-10",
                         savepath: str | None = None, save_format: str = "pdf",
                         figsize: tuple[int | float, int | float] | None = None,
                         to_highlight: list[str] | None = None,
                         head_color: str = constants.COLOR_HEAD_FILL,
                         edge_color: str = constants.COLOR_HEAD_EDGE,
                         fontsize: str | int | float = 10,
                         electrode_color: str = "#ffffff",
                         electrode_highlight_color: str = constants.COLOR_HIGHLIGHT_ELECTRODE) -> None:
    f"""Plot a standard EEG electrode montage.

    :param montage: EEG montage system (currently supporting "10-10" and "10-20")
    {COMMON_PLOT_PARAMS}
    :param to_highlight: Which electrodes to highlight with different color (none will be highlighted by default)
    {MONTAGE_PLOT_PARAMS}
    {MONTAGE_ELECTRODE_COLORS_PARAMS}
    """

    locations = _get_locations_for_montage(montage)
    electrode_names = list(locations.keys())
    electrode_colors = _get_electrode_colors(used_names=electrode_names, 
                                             to_show=electrode_names, 
                                             to_highlight=to_highlight or [],
                                             head_color=head_color, electrode_color=electrode_color, electrode_highlight_color=electrode_highlight_color)

    if figsize is None:
        figsize = (8, 8)
    fig, ax = plt.subplots(figsize=figsize)

    _plot_electrode_montage(
        ax=ax,
        locations=locations, electrode_colors=electrode_colors,
        head_color=head_color, edge_color=edge_color, fontsize=fontsize
    )
    _finalize_plot(savepath=savepath, save_format=save_format)


@enforce_type_hints
def plot_electrode_values(recording: Recording,
                          start: float | int, duration: float | int = 1.0,
                          savepath: str | None = None, save_format: str = "pdf",
                          figsize: tuple[int | float, int | float] | None = None,
                          head_color: str = constants.COLOR_HEAD_FILL,
                          edge_color: str = constants.COLOR_HEAD_EDGE,
                          fontsize: str | int | float = 10,
                          cmap: str = 'coolwarm', vmin: int | float | None = None, vmax: int | float | None = None) -> None:
    f"""Plot electrodes colored by mean signal values of a certain recording moment.

    :param recording: Input EEG recording
    {LENGTH_PLOT_PARAMS}
    {COMMON_PLOT_PARAMS}
    {MONTAGE_PLOT_PARAMS}    
    {HEATMAP_PARAMS}
    """

    montage_name = _detect_montage_type(recording.channel_names)
    locations = _get_locations_for_montage(montage_name)

    if figsize is None:
        figsize = (8, 8)
    fig, ax = plt.subplots(figsize=figsize)

    start_sample, n_samples, _ = _get_timepoints(number_of_samples=recording.number_of_samples, sample_rate=recording.sample_rate, start=start, duration=duration)
    subrecording = cut_by_sample_range(recording, start_sample, start_sample+n_samples)

    values = {}
    for i, name in enumerate(subrecording.channel_names):
        mean_value = subrecording.data[i].mean()
        values[name] = mean_value

    electrode_colors = _values_to_colors(values=values, cmap=cmap, vmin=vmin, vmax=vmax)

    _plot_electrode_montage(
        ax=ax,
        locations=locations, electrode_colors=electrode_colors,
        head_color=head_color, edge_color=edge_color, fontsize=fontsize
    )
    _finalize_plot(savepath=savepath, save_format=save_format)


# -------- topography map plots


@enforce_type_hints
def plot_mean_topography(recording: Recording,
                         start: float | int, duration: float | int = 1.0,
                         savepath: str | None = None, save_format: str = "pdf",
                         figsize: tuple[int | float, int | float] | None = None,
                         head_color: str = constants.COLOR_HEAD_FILL,
                         edge_color: str = constants.COLOR_HEAD_EDGE,
                         fontsize: str | int | float = 10,
                         grid_resolution: int = constants.TOPOGRAPHY_GRID_RESOLUTION, 
                         contour_levels: int = constants.TOPOGRAPHY_CONTOUR_LEVELS,
                         contour_alpha: float | int = constants.TOPOGRAPHY_CONTOUR_ALPHA,
                         contour_color: str = constants.TOPOGRAPHY_CONTOUR_COLOR,
                         contour_width: int | float = constants.TOPOGRAPHY_CONTOUR_WIDTH,
                         show_electrodes: bool = True,
                         electrode_color: str = constants.TOPOGRAPHY_ELECTRODE_COLOR, 
                         electrode_size: int | float = constants.TOPOGRAPHY_ELECTRODE_SIZE,
                         label_electrodes: bool = False,
                         cmap: str = 'RdBu_r', vmin: int | float | None = None, vmax: int | float | None = None) -> None:
    f"""Plot a mean topographic map of a certain recording moment.

    :param recording: Input EEG recording
    {LENGTH_PLOT_PARAMS}
    {COMMON_PLOT_PARAMS}
    {MONTAGE_PLOT_PARAMS}
    {TOPOGRAPHY_PARAMS}
    {HEATMAP_PARAMS}
    """

    montage_name = _detect_montage_type(recording.channel_names)
    locations = _get_locations_for_montage(montage_name, recording.channel_names)

    if figsize is None:
        figsize = (8, 8)
    fig, ax = plt.subplots(figsize=figsize)

    start_sample, n_samples, _ = _get_timepoints(number_of_samples=recording.number_of_samples, sample_rate=recording.sample_rate, start=start, duration=duration)
    subrecording = cut_by_sample_range(recording, start_sample, start_sample+n_samples)

    values = {}
    for i, name in enumerate(subrecording.channel_names):
        mean_value = subrecording.data[i].mean()
        values[name] = mean_value

    _plot_eeg_topography(ax=ax,
                         locations=locations, values=values,
                         grid_resolution=grid_resolution, cmap=cmap, vmin=vmin, vmax=vmax, vcenter=0,
                         contour_levels=contour_levels, contour_alpha=contour_alpha,
                         contour_color=contour_color, contour_width=contour_width,
                         show_electrodes=show_electrodes, electrode_color=electrode_color,
                         electrode_size=electrode_size, label_electrodes=label_electrodes, fontsize=fontsize,
                         head_color=head_color, edge_color=edge_color)
    _finalize_plot(savepath=savepath, save_format=save_format)


@enforce_type_hints
def plot_frequency_topography(recording: Recording, frequency: int,
                              start: float | int = 0, duration: float | int = -1,
                              savepath: str | None = None, save_format: str = "pdf",
                              figsize: tuple[int | float, int | float] | None = None,
                              head_color: str = constants.COLOR_HEAD_FILL,
                              edge_color: str = constants.COLOR_HEAD_EDGE,
                              fontsize: str | int | float = 10,
                              grid_resolution: int = constants.TOPOGRAPHY_GRID_RESOLUTION, 
                              contour_levels: int = constants.TOPOGRAPHY_CONTOUR_LEVELS,
                              contour_alpha: float | int = constants.TOPOGRAPHY_CONTOUR_ALPHA,
                              contour_color: str = constants.TOPOGRAPHY_CONTOUR_COLOR,
                              contour_width: int | float = constants.TOPOGRAPHY_CONTOUR_WIDTH,
                              show_electrodes: bool = True,
                              electrode_color: str = constants.TOPOGRAPHY_ELECTRODE_COLOR, 
                              electrode_size: int | float = constants.TOPOGRAPHY_ELECTRODE_SIZE,
                              label_electrodes: bool = False,
                              cmap: str = 'magma', vmin: float | None = None, vmax: float | None = None) -> None:
    """Plot a single-frequency power topographic map of a certain recording moment.

    :param recording: Input EEG recording
    :param frequency: Target frequency (in Hz)
    {LENGTH_PLOT_PARAMS}
    {COMMON_PLOT_PARAMS}
    {MONTAGE_PLOT_PARAMS}
    {TOPOGRAPHY_PARAMS}
    {HEATMAP_PARAMS}
    """

    montage_name = _detect_montage_type(recording.channel_names)
    locations = _get_locations_for_montage(montage_name, recording.channel_names)

    if figsize is None:
        figsize = (8, 8)
    fig, ax = plt.subplots(figsize=figsize)

    if (start != 0) or (duration != -1):
        start_sample, n_samples, _ = _get_timepoints(number_of_samples=recording.number_of_samples, sample_rate=recording.sample_rate, 
                                                        start=start, duration=duration)
        subrecording = cut_by_sample_range(recording, start_sample, start_sample+n_samples)
    else:
        subrecording = recording

    f, periodograms = get_frequency_spectrum(subrecording)
    freq_i = np.searchsorted(f, frequency)
    values = {ch:v for ch, v in zip(recording.channel_names, periodograms[:, freq_i])}

    _plot_eeg_topography(ax=ax,
                         locations=locations, values=values,
                         grid_resolution=grid_resolution, cmap=cmap, vmin=vmin, vmax=vmax, vcenter=None,
                         contour_levels=contour_levels, contour_alpha=contour_alpha,
                         contour_color=contour_color, contour_width=contour_width,
                         show_electrodes=show_electrodes, electrode_color=electrode_color,
                         electrode_size=electrode_size, label_electrodes=label_electrodes, fontsize=fontsize,
                         head_color=head_color, edge_color=edge_color)
    _finalize_plot(savepath=savepath, save_format=save_format)


# -------- ICA


def plot_ica_components(recording: Recording, n_components: int, duration: int = 10,
                        trim_start: float | int = 1.0, trim_end: float | int = 1.0,
                        normal_range: float | int = 5.0, random_state: int = 42) -> None:
    """Plot Independent Component Analysis (ICA) components from EEG data.

    :param recording: Input EEG recording
    :param n_components: Number of ICA components to compute and display
    :param duration: Duration of signal segments to show around min/max values (in seconds)
    :param trim_start: Seconds to exclude from the beginning of the recording (to avoid boundary artifacts)
    :param trim_end: Seconds to exclude from the end of the recording (to avoid boundary artifacts)
    :param normal_range: Y-axis range for reference lines
    :param random_state: Random seed for reproducible ICA decomposition
    """

    # Trim boundary artifacts if requested
    if trim_start > 0 or trim_end > 0:
        if trim_start + trim_end >= recording.duration:
            raise BreesyError(
                f"Cannot trim {trim_start}s from start and {trim_end}s from end "
                f"of a {recording.duration}s recording.",
                f"Make sure that trim_start + trim_end < recording duration. "
                f"Try reducing trim_start or trim_end parameters, or use a longer recording."
            )
        trimmed_recording = cut_by_second_range(
            recording=recording,
            start_second=trim_start,
            end_second=recording.duration - trim_end
        )
    else:
        trimmed_recording = recording

    components = get_ica_components(recording=trimmed_recording, n_components=n_components, random_state=random_state)
    time = np.arange(0, trimmed_recording.number_of_samples) / trimmed_recording.sample_rate

    min_signal = np.abs(components).argmin(axis=1)
    max_signal = np.abs(components).argmax(axis=1)
    halflen = int(duration // 2 * trimmed_recording.sample_rate) 

    fig, ax = plt.subplots(n_components, 2, figsize=(14, n_components*1))
    component_iter = tqdm(range(n_components), desc="Plotting ICA components", disable=n_components < 2)
    for i in component_iter:
        _plot_ica_component_segment(
            ax=ax[i, 0],
            time=time, component=components[i],
            sample=min_signal[i], halflen=halflen, normal_range=normal_range)
        _plot_ica_component_segment(
            ax=ax[i, 1],
            time=time, component=components[i],
            sample=max_signal[i], halflen=halflen, normal_range=normal_range)
        ax[i, 0].set_ylabel(f'IC {i}')        

    plt.suptitle('ICA sources')
    ax[0, 0].set_title("Around min source value")
    ax[0, 1].set_title("Around max source value")
        
    plt.tight_layout()
    plt.show()


def _plot_ica_component_segment(ax, time: np.ndarray, component: np.ndarray,
                                sample: int, halflen: int, normal_range: float) -> None:
    help_line_kwargs = {'alpha': 0.6, 'lw': 1, 'ls': '--', 'c': '#ff8800'}
    center_line_kwargs = {'alpha': 0.6, 'lw': 1, 'ls': '--', 'c': '#888888'}
    start, end = max(0, sample - halflen), min(sample + halflen, component.shape[-1])
    ax.plot(time[start:end], component[start:end], lw=1)
    ax.axhline(-normal_range, **help_line_kwargs)
    ax.axhline(normal_range, **help_line_kwargs)
    ax.axhline(0, **center_line_kwargs)
    ax.axvline(0, **help_line_kwargs)
    ax.axvline(time[-1], **help_line_kwargs)
    ax.set_xlim(time[start], time[end])
    ax.set_ylim(component[start:end].min(), component[start:end].max())


# -------- internal functions


def _plot_signal(ax: Axes,
                 data: np.ndarray, number_of_samples: int,
                 sample_rate: int | float, channel_names: list[str],
                 start: float | int, duration: float | int,
                 scaling: int | float, cmap: str, cmap_size: int,
                 center_data: bool = True) -> None:
    
    n_channels = len(channel_names)
    start_sample, n_samples, time = _get_timepoints(number_of_samples=number_of_samples, sample_rate=sample_rate, start=start, duration=duration)

    visible_data = data[:, start_sample:n_samples+start_sample]
    offset = max([visible_data[i].std() * 2 for i in range(n_channels)]) / scaling

    color_cycle = _get_color_cycle(cmap=cmap, cmap_size=cmap_size)
    plt.box(False)

    yticks_positions = []
    for i in range(n_channels):
        to_plot = visible_data[i]
        if center_data:
            to_plot = to_plot - to_plot.mean() + i * offset
        ax.plot(time, to_plot, lw=1, alpha=0.8, color=next(color_cycle))
        yticks_positions.append(to_plot[0])

    ax.set_yticks(yticks_positions, channel_names)
    ax.set_ylabel('Channel name')
    ax.set_xlabel('Time (s)')
    ax.set_xlim(start, start+duration)
    ax.grid(True)


def _plot_electrode_montage(ax: Axes,
                            locations: MutableMapping[str, tuple[float, float]], electrode_colors: dict[str, str],
                            head_color: str, edge_color: str,
                            fontsize: str | int | float) -> None:
    
    not_in_locations = set(electrode_colors.keys()) - set(locations.keys())
    if not_in_locations:
        logger.warning(f'Unrecognized channel names: {", ".join(list(not_in_locations))}. Cannot obtain locations for these channels, so will ignore them.')

    _draw_head(ax=ax, head_color=head_color, edge_color=edge_color)

    for name, location in locations.items():
        if name in electrode_colors:
            _draw_electrode(ax=ax, name=name, location=location, color=electrode_colors[name], edge_color=edge_color, fontsize=fontsize)

    ax.set_aspect('equal')
    ax.set_axis_off()


def _draw_head(ax: Axes, head_color: str, edge_color: str) -> None:
    import matplotlib.patches as patches

    head_radius = 1
    ear_left = patches.Ellipse(
        xy=(-head_radius, 0),
        width=head_radius / 6,
        height=head_radius / 3,
        fill=True,
        facecolor=head_color,
        edgecolor=edge_color,
        lw=2
    )
    ear_right = patches.Ellipse(
        xy=(head_radius, 0),
        width=head_radius / 6,
        height=head_radius / 3,
        fill=True,
        facecolor=head_color,
        edgecolor=edge_color,
        lw=2
    )
    nose = patches.FancyArrow(
        x=0,
        y=head_radius * 0.95,
        dx=0,
        dy=head_radius / 5,
        fill=True,
        facecolor=head_color,
        edgecolor=edge_color,
        head_width=head_radius / 3,
        head_length=head_radius / 5,
        length_includes_head=True,
        lw=2
    )
    head_circle = patches.Circle(
        xy=(0, 0),
        radius=head_radius,
        facecolor=head_color,
        edgecolor=edge_color,
        fill=True,
        lw=2
    )
    inner_circle = patches.Circle(
        xy=(0, 0),
        radius=head_radius * 0.8,
        color=edge_color,
        fill=False,
        ls='--',
        lw=1
    )

    ax.add_patch(ear_left)
    ax.add_patch(ear_right)
    ax.add_patch(nose)
    ax.add_patch(head_circle)
    ax.add_patch(inner_circle)
    ax.plot([-head_radius, head_radius], [0, 0], ls='--', lw=1, color=edge_color)
    ax.plot([0, 0], [-head_radius, head_radius], ls='--', lw=1, color=edge_color)


def _draw_electrode(ax: Axes, name: str, location: tuple[float, float],
                    color: str, edge_color: str, fontsize: str | int | float) -> None:
    import matplotlib.patches as patches

    electrode_circle = patches.Circle(
        xy=location,
        radius=1 / 14,
        facecolor=color,
        edgecolor=edge_color,
        fill=True,
        zorder=2
    )
    ax.add_patch(electrode_circle)

    ax.text(
        x=location[0],
        y=location[1],
        s=name,
        color=edge_color,
        va='center',
        ha='center',
        fontsize=fontsize,
        zorder=2
    )


def _interpolated_head_grid(coords_arr: np.ndarray, values_arr: np.ndarray, grid_resolution: int):
    grid_extent = 1.1  # Slightly larger than head
    xi = np.linspace(-grid_extent, grid_extent, grid_resolution)
    yi = np.linspace(-grid_extent, grid_extent, grid_resolution)
    xi_grid, yi_grid = np.meshgrid(xi, yi)
    
    rbf = _si_get_rbf_interpolator(coords_arr, values_arr)
    grid_points = np.column_stack([xi_grid.ravel(), yi_grid.ravel()])
    interpolated_values = rbf(grid_points).reshape(xi_grid.shape)

    distance_from_center = np.sqrt(xi_grid**2 + yi_grid**2)
    circular_mask = distance_from_center <= 1
    masked_values = np.where(circular_mask, interpolated_values, np.nan)

    return xi_grid, yi_grid, masked_values


def _plot_eeg_topography(ax: Axes,
                         locations: MutableMapping[str, tuple[float, float]], values: dict[str, float],
                         grid_resolution: int, cmap: str, vmin: int | float | None, vmax: int | float | None, vcenter: int | float | None,
                         contour_levels: int, contour_alpha: float | int, contour_color: str, contour_width: int | float,
                         show_electrodes: bool, electrode_color: str,  electrode_size: float | int, label_electrodes: bool,
                         fontsize: str | int | float, head_color: str, edge_color: str) -> None:
    import matplotlib.patches as patches

    _draw_head(ax=ax, head_color=head_color, edge_color=edge_color)

    ch_names = list(locations.keys())
    coords_arr = np.array([locations[ch] for ch in ch_names])
    values_arr = np.array([values[ch] for ch in ch_names])
    xi_grid, yi_grid, interpolated = _interpolated_head_grid(
        coords_arr=coords_arr, values_arr=values_arr, grid_resolution=grid_resolution
    )

    norm = _get_two_slope_norm(data=interpolated, vmin=vmin, vmax=vmax, vcenter=vcenter)
    contour_fill = ax.contourf(xi_grid, yi_grid, interpolated, 
                               levels=contour_levels, cmap=cmap, norm=norm,
                               extend='both')
    ax.contour(xi_grid, yi_grid, interpolated, 
               levels=contour_levels, colors=contour_color, 
               alpha=contour_alpha, linewidths=contour_width)

    if show_electrodes:
        ax.scatter(coords_arr[:, 0], coords_arr[:, 1], c=electrode_color, s=electrode_size, zorder=5)
        if label_electrodes:
            for name in ch_names:
                ax.annotate(name, locations[name], fontsize=fontsize)

    head_circle = patches.Circle(xy=(0, 0), radius=1, color=edge_color, fill=False, lw=2)
    ax.add_patch(head_circle)

    ax.set_aspect('equal')
    ax.set_axis_off()

    cbar = plt.colorbar(contour_fill, ax=ax, shrink=0.8, aspect=20)
    cbar.set_label('Amplitude', rotation=270, labelpad=15)


def _finalize_plot(savepath: str | None, save_format: str) -> None:
    plt.tight_layout()
    if savepath:
        plt.savefig(savepath, format=save_format)
    plt.show()


# ---- helping functions


def _get_timepoints(number_of_samples: int, sample_rate: int | float, start: int | float, duration: int | float) -> tuple[int, int, np.ndarray]:
    start_sample = int(start * sample_rate)
    if duration == -1:
        duration = number_of_samples - start_sample
    n_samples = int(duration * sample_rate)
    time = np.arange(n_samples) / sample_rate + start
    return start_sample, n_samples, time


def _get_color_cycle(cmap: str = 'Dark2', cmap_size: int = 8):
    from matplotlib import colormaps
    from itertools import cycle
    return cycle(colormaps[cmap](np.linspace(0, 1, cmap_size)))


def _get_electrode_colors(used_names: list[str], to_show: list[str], to_highlight: list[str],
                          head_color: str, electrode_color: str, electrode_highlight_color: str) -> dict[str, str]:
    colors = {}
    for name in to_show:
        if name in to_highlight:
            color = electrode_highlight_color
        elif name not in used_names:
            color = head_color
        else:
            color = electrode_color
        colors[name] = color
    return colors


def _values_to_colors(values: dict[str, float], cmap: str, vmin: int | float | None, vmax: int | float | None) -> dict:
    import matplotlib.colors as colors
    from matplotlib import colormaps
    vmin = vmin or min(values.values())
    vmax = vmax or max(values.values())
    norm = colors.Normalize(vmin=vmin, vmax=vmax)
    colormap = colormaps.get_cmap(cmap)
    color_dict = {name: colormap(norm(values[name])) for name in values}
    return color_dict


def _get_two_slope_norm(data: np.ndarray, vmin: int | float | None, vmax: int | float | None, vcenter: int | float | None):
    from matplotlib.colors import TwoSlopeNorm
    v_min = vmin if (vmin is not None) else np.nanmin(data)
    v_max = vmax if (vmax is not None) else np.nanmax(data)

    if (vcenter is None) or (vcenter <= v_min) or (vcenter >= v_max):
        v_center = (v_min + v_max) / 2
    else:
        v_center = vcenter

    norm = TwoSlopeNorm(vcenter=v_center, vmin=v_min, vmax=v_max)
    return norm


def _get_locations_for_montage(montage_name: str = "10-10", channel_names: list[str] | None = None) -> MutableMapping[str, tuple[float, float]]:
    if montage_name in ['10-20', '10_20', '1020']:
        locations = constants.EEG_MONTAGES['10-20']
    elif montage_name in ['10-10', '10_10', '1010']:
        locations = constants.EEG_MONTAGES['10-10']
    else:
        raise BreesyError(f'Unknown montage type: {montage_name}. Use either 10-10 (default) or 10-20.')
    if channel_names is not None:
        locations = {ch: coords for ch, coords in locations.items() if ch in channel_names}
        if len(locations) < len(channel_names):
            unknown_channels = [ch for ch in channel_names if ch not in locations.keys()]
            raise BreesyError(f'Found unknown channel(s) not defined for the {montage_name} montage: {", ".join(unknown_channels)}')
    return locations


def _detect_montage_type(ch_names: list[str]) -> str:
    if any([ch.lower() in [x.lower() for x in constants.EEG_MONTAGES['10-10'].keys()] for ch in ch_names]):
        logger.info("Using 10-10 montage...")
        return "10-10"
    else:
        logger.info("Using 10-20 montage...")
        return "10-20"



import logging
import sys

logger = logging.getLogger("breesy")
logger.setLevel(logging.INFO)

# Prevent bubbling of messages to root logger (avoid duplicates)
logger.propagate = False

_handler = logging.StreamHandler(sys.stdout)
_handler.setFormatter(logging.Formatter("%(levelname)s: %(message)s"))  # e.g.: logger.warning("hello") -> "WARNING: hello"
logger.addHandler(_handler)

def set_log_level(level: str = "INFO") -> None:
    """Set logging verbosity level for the whole Breesy package.

    :param level: One of 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'.
    """
    numeric_level = getattr(logging, level.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError(f"Invalid log level: {level}. Use one of: DEBUG, INFO, WARNING, ERROR, CRITICAL")
    logger.setLevel(numeric_level)


def silence_logs() -> None:
    """Silence all Breesy log messages."""
    logger.setLevel(logging.CRITICAL + 1)  # Higher than any level


def enable_logs(level: str = "INFO") -> None:
    """Re-enable Breesy log messages after silencing.

    :param level: The log level to restore. Default is 'INFO'.
    """
    set_log_level(level)


from pathlib import Path
import numpy as np
import re, unicodedata

# from .Event import Event
from .recording import Recording, RecordingMetadata
from .errors import BreesyError


ERROR_MESSAGE_CONTACT_THE_AUTHORS = 'Contact the authors of the header (.vhdr) file and ask to fix it.'
ERROR_MESSAGE_OPEN_HEADER_IN_NOTEPAD = 'Also, header (.vhdr) files can be opened and inspected with Notepad or other text editor.'


def load_brainvision_eeg(filename: str) -> Recording:
    """Load a BrainVision EEG file.

    :param filename: Path to a BrainVision file, either a header file (.vhdr) or data file (.eeg, .avg, or .seg)

    :return: Recording object containing the loaded data
    """
    file_path = Path(filename)
    if file_path.suffix.lower() in ['.eeg', '.avg', '.seg']:
        vhdr_path = file_path.with_suffix('.vhdr')
        eeg_path = file_path
    elif file_path.suffix.lower() == '.vhdr':
        vhdr_path = file_path
        eeg_path = None

    if not vhdr_path.exists():
        raise BreesyError(
            f"Header file not found for {filename}. Breesy searched for: {vhdr_path}.",
            "BrainVision files require a .vhdr file. Make sure to either provide a correct .vhdr filename, or a filename for a data file with same name and just a different extension."
        )

    metadata = _parse_brainvision_header(vhdr_path)
    if eeg_path is None:
        eeg_path = vhdr_path.parent / metadata['DataFile']
    if not eeg_path.exists():
        raise BreesyError(
            f"Header file references a daat file but it was not found.",
            f"Make sure the data file exists according to the path provided in the header file, which is:\n{metadata['DataFile']}\nUsually, the data file should be in the same directory as the .vhdr file. "
            f'{ERROR_MESSAGE_OPEN_HEADER_IN_NOTEPAD} You can manually change the name of the data file inside it (at your own risk!), using the "DataFile" parameter.'
        )

    data = _read_brainvision_data(eeg_path, metadata)

    # TODO: finish vmrk function
    # vmrk_path = vhdr_path.parent / metadata['MarkerFile']
    # events = _parse_brainvision_markers(vmrk_path, metadata['SamplingInterval'])
    # if not events:
    #     print("No events found in marker file or marker file not present")

    recording_metadata = RecordingMetadata(
        name=vhdr_path.stem,
        file_path=vhdr_path,
        description=f"BrainVision data loaded from {filename}"
    )

    return Recording(
        data=data,
        channel_names=metadata['ch_names'],
        sample_rate=1_000_000 // int(metadata['SamplingInterval']),  # micros to Hz
        # events=events,
        metadata=recording_metadata
    )


def _find_config_parameter_value(config: str, parameter_name: str, mandatory: bool = False) -> str | None:
    values = (x for x in config.split('\n') if x.startswith(parameter_name))
    try:
        value = next(values)
        try:
            next(values)
            raise BreesyError(f'Incorrectly formatted header file: found two values for parameter "{parameter_name}" in same section.',
                              f'{ERROR_MESSAGE_CONTACT_THE_AUTHORS} {ERROR_MESSAGE_OPEN_HEADER_IN_NOTEPAD} '
                              f'If you indeed find more than one mention of parameter "{parameter_name}", you can manually delete one of it (at your own risk!).')
        except StopIteration:
            return value.split(parameter_name+'=')[1]
    except StopIteration:
        if mandatory:
            raise BreesyError(f'Incorrectly formatted header file: mandatory parameter "{parameter_name}" not found.', 
                              f'{ERROR_MESSAGE_CONTACT_THE_AUTHORS} {ERROR_MESSAGE_OPEN_HEADER_IN_NOTEPAD} '
                              f'If you indeed cannot find a row starting with "{parameter_name}", you can manually add that row (at your own risk!).')
        return None


def _get_binary_dtype(binary_format: str):    
    if binary_format == 'INT_16':
        return np.int16
    elif binary_format == 'INT_32':
        return np.int32
    elif binary_format == 'IEEE_FLOAT_32':
        return np.float32
    raise BreesyError(f'Unknown binary format specified: "{binary_format}".',
                      f'{ERROR_MESSAGE_CONTACT_THE_AUTHORS} {ERROR_MESSAGE_OPEN_HEADER_IN_NOTEPAD} '
                       'In the "BinaryFormat" field, BrainVision files only allow "IEEE_FLOAT_32" or "INT_16", and Breesy also allows "INT_32".')


def _reshape_data_by_orientation(raw: np.ndarray, orientation: str, n_channels: int) -> np.ndarray:
    n_samples = len(raw) // n_channels
    if orientation == 'MULTIPLEXED':  # Data is interleaved: Ch1_S1, Ch2_S1, ..., ChN_S1, Ch1_S2, ...
        return raw.reshape((n_samples, n_channels)).T
    elif orientation == 'VECTORIZED':  # Data is by channel: Ch1_S1, Ch1_S2, ..., Ch1_SN, Ch2_S1, ...
        return raw.reshape((n_channels, n_samples))
    raise BreesyError(f'Unknown file orientation specified: "{orientation}".',
                      f'{ERROR_MESSAGE_CONTACT_THE_AUTHORS} {ERROR_MESSAGE_OPEN_HEADER_IN_NOTEPAD} '
                       'In the "DataOrientation" field, BrainVision files only allow "MULTIPLEXED", and Breesy also allows "VECTORIZED".')


def _get_resolutions_according_to_units(scales: list[float], units: list[str]) -> list[float]:
    resolutions = []
    for i, (scale, unit) in enumerate(zip(scales, units)):
        unit_normalized = unit.lower()
        if unit_normalized == 'μv':
            resolutions.append(scale)
        elif unit_normalized == 'mv':
            resolutions.append(scale * 1000)
        elif unit_normalized == 'v':
            resolutions.append(scale * 1_000_000)
        else:
            raise BreesyError(f'Unknown unit specified for channel nr. {i+1}: "{unit}".',
                              f'{ERROR_MESSAGE_CONTACT_THE_AUTHORS} Breesy only supports "μV", "mV", and "V" (not case-sensitive). '
                              f'If you think that "{unit}" unit should also be supported, contact Breesy authors.')
    return resolutions



def _read_brainvision_data(eeg_path: Path, metadata: dict) -> np.ndarray:
    dtype = _get_binary_dtype(binary_format=metadata['BinaryFormat'])   
    with open(eeg_path, 'rb') as f:
        raw_data = np.fromfile(f, dtype=dtype)
    data = _reshape_data_by_orientation(
        raw=raw_data,
        orientation=metadata['DataOrientation'],
        n_channels=int(metadata['NumberOfChannels'])
    )
    resolutions = _get_resolutions_according_to_units(scales=metadata['ch_scales'], units=metadata['ch_units'])
    return data.astype(np.float64) * np.array(resolutions).reshape(-1, 1)


def _parse_brainvision_header(vhdr_path: Path) -> dict:
    # Official specifications: https://www.brainproducts.com/support-resources/brainvision-core-data-format-1-0/
    with open(vhdr_path, 'r', encoding='utf-8') as f:
        config = unicodedata.normalize('NFKC', f.read())
        
    if (
        not (config.startswith("BrainVision Data Exchange Header File Version 1.0") or 
             config.startswith("Brain Vision Data Exchange Header File Version 1.0")) or
        '[Common Infos]' not in config or
        '[Channel Infos]' not in config
    ):
        raise BreesyError(f'Provided file is not valid according to the specification of BrainVision Core Data Format 1.0.',
                          f'Check that the path provided is a correct header file (ending with ".vhdr"):\n{vhdr_path}\n{ERROR_MESSAGE_CONTACT_THE_AUTHORS}')
        
    vhdr = {}
    sections = re.findall(r'(\[.+\]\n[\s\S]+?(?:\n\n|\Z))', config)

    for section in sections:

        if section.startswith('[Common Infos]'):
            for key in ['DataFile', 'DataOrientation', 'NumberOfChannels', 'SamplingInterval']:
                vhdr[key] = _find_config_parameter_value(section, key, mandatory=True)
            for key in ['MarkerFile', 'Averaged']:
                vhdr[key] = _find_config_parameter_value(section, key, mandatory=False)

        elif section.startswith('[Binary Infos]'):
            vhdr['BinaryFormat'] = _find_config_parameter_value(section, 'BinaryFormat', mandatory=True)

        elif section.startswith('[Channel Infos]'):
            vhdr['ch_names'] = []
            vhdr['ch_refs'] = []
            vhdr['ch_scales'] = []
            vhdr['ch_units'] = []
            for row in section.split('\n'):
                if row.startswith(';'):
                    continue
                if row.startswith('Ch'):
                    ch_i, ch_params = row.split('=')
                    expected_ch_name = f"Ch{len(vhdr['ch_names']) + 1}"
                    if ch_i != expected_ch_name:
                        raise BreesyError(f'Channel named "{ch_i}" is given at position which should be channel named "{expected_ch_name}".',
                                          f'Information about some channels is either missed or repeated. {ERROR_MESSAGE_CONTACT_THE_AUTHORS}')

                    ch_params = ch_params.split(',')
                    vhdr['ch_names'].append(ch_params[0])
                    vhdr['ch_refs'].append(ch_params[1])
                    vhdr['ch_scales'].append(float(ch_params[2]))
                    vhdr['ch_units'].append(ch_params[3] if len(ch_params) == 4 else 'μV')
                    
    return vhdr


from pathlib import Path
import numpy as np
from breesy.load import load_recording
from breesy.features import get_mean_power
from breesy.recording import split_by_window_duration

REPO_ROOT = Path(__file__).parent.parent


def test_mean_power_alpha_band():
    # Arrange
    datapath = REPO_ROOT / "tests" / "data" / "alpha" / "subject_01.mat"
    rec = load_recording(str(datapath))
    epochs = split_by_window_duration(rec, window_duration=5.0)
    epoch_data = np.array([epoch.data for epoch in epochs])

    # Act
    alpha_power = get_mean_power(epoch_data, freq_range='alpha', sample_rate=rec.sample_rate)

    # Assert
    assert alpha_power.ndim == 1
    assert alpha_power.shape[0] == len(epochs)
    assert np.all(alpha_power > 0)
    assert np.all(np.isfinite(alpha_power))


def test_mean_power_custom_range():
    # Arrange
    datapath = REPO_ROOT / "tests" / "data" / "alpha" / "subject_01.mat"
    rec = load_recording(str(datapath))
    epochs = split_by_window_duration(rec, window_duration=5.0)
    epoch_data = np.array([epoch.data for epoch in epochs])

    # Act
    power = get_mean_power(epoch_data, freq_range=(10, 15), sample_rate=rec.sample_rate)

    # Assert
    assert power.ndim == 1
    assert power.shape[0] == len(epochs)
    assert np.all(power > 0)
    assert np.all(np.isfinite(power))


def test_frequency_bands():
    # Arrange
    datapath = REPO_ROOT / "tests" / "data" / "alpha" / "subject_01.mat"
    rec = load_recording(str(datapath))
    epochs = split_by_window_duration(rec, window_duration=5.0)
    epoch_data = np.array([epoch.data for epoch in epochs])
    bands = ['delta', 'theta', 'alpha', 'beta', 'gamma']

    # Act & Assert
    for band in bands:
        power = get_mean_power(epoch_data, freq_range=band, sample_rate=rec.sample_rate)

        assert power.ndim == 1
        assert power.shape[0] == len(epochs)
        assert np.all(power > 0)
        assert np.all(np.isfinite(power))



from urllib.request import urlopen, Request
import certifi, ssl
from shutil import copyfileobj
from pathlib import Path
from tqdm.auto import tqdm

from .errors import BreesyError
from .log import logger


AVAILABLE_DATAFILES = ['alpha', 'pangolin']


def download_example_data(dataset_name: str, dir_name: str = "tmp", ssl_context=None) -> str:
    f"""Download example EEG file for testing and learning.

    Downloads publicly available EEG data files. Currently available files include: {AVAILABLE_DATAFILES}.

    :param dataset_name: Name of the file to load
    :param dir_name: Directory to save downloaded file (will be created if does not exist)
    :param ssl_context: SSL context for secure downloads

    :return: path to the downloaded file
    """

    # TODO: some downloads take time, do some loading bar?
    # TODO: add urls for more information about each dataset
    if dataset_name == "alpha":
        filename = "subject_01.mat"
        url = "https://zenodo.org/records/2348892/files/subject_01.mat?download=1"
        # ch_names = ["Fp1", "Fp2", "FC5", "FC6", "Fz", "T7", "Cz", "T8",
        #             "P7", "P3", "Pz", "P4", "P8", "O1", "Oz", "O2"]
        # sample_rate = 512
    
    # elif dataset_name == "madhd":
    #     filename = "MADHD.mat"
    #     url = "https://data.mendeley.com/public-files/datasets/6k4g25fhzg/files/9c5928cf-f8ef-485b-ac5b-f17b2df935f3/file_downloaded"
    #     # ch_names = ["Cz", "F4"]
    #     # sample_rate = 256

    elif dataset_name == "pangolin":
        filename = "S1_run8.mat"
        url = "https://osf.io/download/jrcn8/?view_only=d23acfd50655427fbaae381a17cbfbcc"
        # ch_names = [f"ch{i+1}" for i in range(258)]  # this dataset has no standard channel names
        # sample_rate = 600

    # elif dataset_name == "schalk":
    #     filename = "S001R01.edf"
    #     url = "https://www.physionet.org/files/eegmmidb/1.0.0/S001/S001R01.edf?download"
    #     ch_names = ['Fc5', 'Fc3', 'Fc1', 'Fcz', 'Fc2', 'Fc4', 'Fc6',
    #                 'C5', 'C3', 'C1', 'Cz', 'C2', 'C4', 'C6',
    #                 'Cp5', 'Cp3', 'Cp1', 'Cpz', 'Cp2', 'Cp4', 'Cp6',
    #                 'Fp1', 'Fpz', 'Fp2', 'Af7', 'Af3', 'Afz', 'Af4', 'Af8',
    #                 'F7', 'F5', 'F3', 'F1', 'Fz', 'F2', 'F4', 'F6', 'F8',
    #                 'Ft7', 'Ft8', 'T7', 'T8', 'T9', 'T10', 'Tp7', 'Tp8',
    #                 'P7', 'P5', 'P3', 'P1', 'Pz', 'P2', 'P4', 'P6', 'P8',
    #                 'Po7', 'Po3', 'Poz', 'Po4', 'Po8', 'O1', 'Oz', 'O2', 'Iz'
    #                 ]
    #     sample_rate = 160

    # elif dataset_name == "telemetry":
    #     filename = "ST7011J0-PSG.edf"
    #     url = "https://www.physionet.org/files/sleep-edfx/1.0.0/sleep-telemetry/ST7011J0-PSG.edf?download"
    #     ch_names = range(5)  # TODO: get actual channel names
    #     sample_rate = 100

    else:
        available_string = ", ".join([f'"{n}"' for n in AVAILABLE_DATAFILES])
        raise BreesyError(f'Unknown name of filename provided: "{dataset_name}".', 
                          f'Please use one of available names ({available_string}).')   

    dirpath = Path(dir_name)
    dirpath.mkdir(parents=True, exist_ok=True)
    filepath = dirpath / filename

    if not filepath.exists():
        logger.info(f'Will download the data file from {url}')
        if ssl_context is None:
            ssl_context = ssl.create_default_context(cafile=certifi.where())
        req = Request(url, headers={'User-Agent' : "Magic Browser"})
        with urlopen(req, context=ssl_context) as in_stream:
            file_size = int(in_stream.headers.get('Content-Length', 0))
            with tqdm.wrapattr(in_stream, "read", total=file_size,
                              desc=f"Downloading {filename}",
                              unit='B', unit_scale=True, unit_divisor=1024) as wrapped_stream:
                with open(filepath, 'wb') as out_file:
                    copyfileobj(wrapped_stream, out_file)
        logger.info(f'Done! File saved in: {filepath}')
    else:
        logger.info(f'File already exists: {filepath}')

    return filepath



from pathlib import Path
import numpy as np
from breesy.load import load_recording
from breesy.filtering import notch_filter, select_bandwidth

REPO_ROOT = Path(__file__).parent.parent


def test_notch_filter_removes_powerline():
    # Arrange
    datapath = REPO_ROOT / "tests" / "data" / "alpha" / "subject_01.mat"
    rec = load_recording(str(datapath))

    # Act
    filtered = notch_filter(rec, frequency=50)

    # Assert
    assert filtered.data.shape == rec.data.shape
    assert filtered.sample_rate == rec.sample_rate
    assert filtered.channel_names == rec.channel_names
    assert not np.array_equal(filtered.data, rec.data)


def test_bandpass_filter_applies():
    # Arrange
    datapath = REPO_ROOT / "tests" / "data" / "alpha" / "subject_01.mat"
    rec = load_recording(str(datapath))

    # Act
    filtered = select_bandwidth(rec, low=8, high=13)

    # Assert
    assert filtered.data.shape == rec.data.shape
    assert filtered.sample_rate == rec.sample_rate
    assert filtered.channel_names == rec.channel_names
    assert not np.array_equal(filtered.data, rec.data)


def test_chained_filters():
    # Arrange
    datapath = REPO_ROOT / "tests" / "data" / "alpha" / "subject_01.mat"
    rec = load_recording(str(datapath))

    # Act
    filtered = notch_filter(rec, frequency=50)
    filtered = select_bandwidth(filtered, low=1, high=40)

    # Assert
    assert filtered.data.shape == rec.data.shape
    assert filtered.sample_rate == rec.sample_rate
    assert not np.array_equal(filtered.data, rec.data)
